from __future__ import absolute_import
from __future__ import print_function

from collections import OrderedDict
from math import log, ceil

import veriloggen.core.vtypes as vtypes
import veriloggen.types.fixed as fx
import veriloggen.types.rom as rom
import veriloggen.types.ram as ram
from veriloggen.seq.seq import make_condition as _make_condition

from . import mul
from . import div
from . import madd


# Object ID counter for object sorting key
_object_counter = 0


def Constant(value, fixed=True, point=0):
    if isinstance(value, int):
        return Int(value)

    if isinstance(value, bool):
        v = 1 if value else 0
        return Int(v)

    if isinstance(value, float):
        if fixed:
            value = fx.to_fixed(value, point)
            return FixedPoint(value, point)
        return Float(value)

    if isinstance(value, str):
        return Str(value)

    raise TypeError("Unsupported type for Constant '%s'" % str(type(value)))


def Variable(data=None, width=32, point=0, signed=True):
    return _Variable(data, width, point, signed)


def Parameter(name, value, width=32, point=0, signed=True):
    """ parameter with an immediate value """
    if not isinstance(name, str):
        raise TypeError("'name' must be str, not '%s'" % str(tyep(name)))
    return _ParameterVariable(name, width, point, signed, value=value)


def ParameterVariable(data, width=32, point=0, signed=True):
    """ parameter with an existing object """
    if isinstance(data, float):
        return Constant(data, point=point)
    if isinstance(data, (int, bool)):
        data = vtypes.Int(data, width=width)
    return _ParameterVariable(data, width, point, signed)


class _Node(object):

    def __init__(self):
        global _object_counter
        self.object_id = _object_counter
        _object_counter += 1

        # for graph visualizer
        self.graph_label = None
        self.graph_shape = 'circle'
        self.graph_color = 'black'
        self.graph_style = ''
        self.graph_peripheries = 1

    def __hash__(self):
        object_id = self.object_id if hasattr(self, 'object_id') else None
        return hash((id(self), object_id))

    def __eq__(self, other):
        return (id(self), self.object_id) == (id(other), other.object_id)

    def name(self, prefix=None):
        clsname = self.__class__.__name__.lower()
        if prefix is None:
            prefix = 'tmp'
        return '_'.join(['', clsname, prefix, str(self.object_id)])


class _Numeric(_Node):
    latency = 0

    def __hash__(self):
        object_id = self.object_id if hasattr(self, 'object_id') else None
        return hash((id(self), object_id))

    def __init__(self):
        _Node.__init__(self)

        # set up by _set_managers()
        self.m = None
        self.strm = None
        self.seq = None

        self.output_data = None
        self.output_sig_data = None

        self.output_node = None

        self.sig_data = None

        self.start_stage = None
        self.end_stage = None
        self.sink = []

        # set up by _set_attributes()
        self.width = None
        self.point = None
        self.signed = False

        # stage numbers incremented
        self.delayed_value = OrderedDict()

        # stage numbers NOT incremented
        self.previous_value = OrderedDict()

    def output(self, data):
        if self.output_data is not None:
            raise ValueError('output_data is already assigned.')
        self.output_data = data

        if self.strm is not None:
            self.strm.add(self)

    def output_tmp(self):
        if self.m is None:
            raise ValueError("Module information is not set.")

        self.output(self.name('odata'))

    def prev(self, index):
        if index < 0:
            raise ValueError("index must be greater than 0")

        prev = self
        for i in range(index):
            r = self._get_previous_value(i + 1)
            if r is not None:
                prev = r
                continue
            r = _Prev(prev)
            r._set_parent_value(self)
            self._add_previous_value(i + 1, r)
            prev = r

        return prev

    def write(self, wdata, cond=None):
        raise TypeError("Unsupported method.")

    def read(self):
        if self.output_node is not None and id(self) != id(self.output_node):
            return self.output_node.read()

        if self.output_sig_data is None:
            # set default name
            if self.output_data is None:
                self.output_tmp()

            self._implement_output_sig(self.m, self.seq, aswire=True)

        data = self.output_sig_data

        return data

    def get_signed(self):
        return self.signed

    def get_point(self):
        return self.point

    def bit_length(self):
        return self.width

    def eval(self):
        raise NotImplementedError('eval() is not implemented')

    def _set_attributes(self):
        raise NotImplementedError('_set_attributes() is not implemented')

    def _set_managers(self):
        raise NotImplementedError('_set_managers() is not implemented')

    def _set_module(self, m):
        self.m = m

    def _set_strm(self, strm):
        self.strm = strm

    def _set_seq(self, seq):
        self.seq = seq

    def _implement(self, m, seq, svalid=None, senable=None):
        raise NotImplementedError('_implement() is not implemented.')

    def _implement_input(self, m, seq, aswire=False):
        raise NotImplementedError('_implement_input() is not implemented.')

    def _implement_output(self, m, seq, aswire=False):
        if self.end_stage is None:
            self.end_stage = 0

        self._implement_output_sig(m, seq, aswire)
        data = self.output_sig_data

        m.Assign(data(self.sig_data))

    def _implement_output_sig(self, m, seq, aswire=False):
        if self.output_sig_data is not None:
            return

        if self.m is None:
            raise ValueError("Module information is not set.")

        width = self.bit_length()
        signed = self.get_signed()

        type_i = m.Wire if aswire else m.Input
        type_o = m.Wire if aswire else m.Output

        if isinstance(self.output_data, (vtypes.Wire, vtypes.Output)):
            data = self.output_data
            self.output_sig_data = data
        else:
            data = type_o(self.output_data, width=width, signed=signed)
            self.output_sig_data = data

    def _has_output(self):
        if self.output_data is not None:
            return True
        return False

    def _disable_output(self):
        self.output_data = None

    def _disable_output_sig(self):
        self.output_sig_data = None

    def _set_output_node(self, node):
        self.output_node = node

    def _set_start_stage(self, stage):
        self.start_stage = stage

    def _get_start_stage(self):
        return self.start_stage

    def _has_start_stage(self):
        if self.start_stage is None:
            return False
        return True

    def _set_end_stage(self, stage):
        self.end_stage = stage

    def _get_end_stage(self):
        return self.end_stage

    def _has_end_stage(self):
        if self.end_stage is None:
            return False
        return True

    def _add_sink(self, value):
        self.sink.append(value)

    def _add_delayed_value(self, delay, value):
        if delay in self.delayed_value:
            raise ValueError('%d-delayed value is already allocated.' % delay)
        self.delayed_value[delay] = value

    def _get_delayed_value(self, delay):
        if delay not in self.delayed_value:
            return None
        return self.delayed_value[delay]

    def _add_previous_value(self, delay, value):
        if delay in self.delayed_value:
            raise ValueError('%d-delayed value is already allocated.' % delay)
        self.previous_value[delay] = value

    def _get_previous_value(self, delay):
        if delay not in self.previous_value:
            return None
        return self.previous_value[delay]

    def __lt__(self, r):
        return LessThan(self, r)

    def __le__(self, r):
        return LessEq(self, r)

    def __eq__(self, r):
        return Eq(self, r)

    def __ne__(self, r):
        return NotEq(self, r)

    def __ge__(self, r):
        return GreaterEq(self, r)

    def __gt__(self, r):
        return GreaterThan(self, r)

    def __add__(self, r):
        return Plus(self, r)

    def __sub__(self, r):
        return Minus(self, r)

    def __pow__(self, r):
        return Power(self, r)

    def __mul__(self, r):
        return Times(self, r)

    def __div__(self, r):
        return Divide(self, r)

    def __truediv__(self, r):
        return Divide(self, r)

    def __mod__(self, r):
        return Mod(self, r)

    def __and__(self, r):
        return And(self, r)

    def __or__(self, r):
        return Or(self, r)

    def __xor__(self, r):
        return Xor(self, r)

    def __lshift__(self, r):
        return Sll(self, r)

    def __rshift__(self, r):
        return Srl(self, r)

    def __neg__(self):
        return Uminus(self)

    def __pos__(self):
        return Uplus(self)

    def __abs__(self):
        return Abs(self)

    def __getitem__(self, r):
        if isinstance(r, slice):
            size = self.bit_length()

            right = r.start
            if right is None:
                right = 0
            elif isinstance(right, int) and right < 0:
                right = size - abs(right)

            left = r.stop
            if left is None:
                left = size
            elif isinstance(left, int) and left < 0:
                left = size - abs(left)
            left -= 1

            if isinstance(left, int) and left < 0:
                raise ValueError("Illegal slice index: left = %d" % left)

            step = r.step
            if step is None:
                return Slice(self, left, right)
            else:
                if not (isinstance(left, int) and
                        isinstance(right, int) and
                        isinstance(step, int)):
                    raise ValueError(
                        "Slice with step is not supported in Verilog Slice.")

                if step == 0:
                    raise ValueError("Illegal slice step: step = %d" % step)

                values = [Pointer(self, i)
                          for i in range(right, left + 1, step)]
                values.reverse()
                return Cat(*values)

        if isinstance(r, int) and r < 0:
            r = self.bit_length() - abs(r)

        return Pointer(self, r)

    def sra(self, r):  # shift right arithmetically
        return Sra(self, r)

    def repeat(self, times):
        return Repeat(self, times)

    def slice(self, msb, lsb):
        return Slice(self, msb, lsb)

    def __iter__(self):
        self.iter_size = len(self)
        self.iter_count = 0
        return self

    def __next__(self):
        if self.iter_count >= self.iter_size:
            raise StopIteration()

        ret = Pointer(self, self.iter_count)
        self.iter_count += 1
        return ret

    # for Python2
    def next(self):
        return self.__next__()

    def __len__(self):
        ret = self.bit_length()
        if not isinstance(ret, int):
            raise ValueError("Non int length.")
        return ret

    @property
    def raw_data(self):
        if self.sig_data is None:
            raise ValueError(
                "Stream is not synthesized yet. Run Stream.implement().")
        return self.sig_data

    @property
    def data(self):
        if self.output_node is not None:
            return self.output_node.output_sig_data
        return self.raw_data


class _Operator(_Numeric):
    latency = 1

    def _implement(self, m, seq, svalid=None, senable=None):
        raise NotImplementedError('_implement() is not implemented.')


class _BinaryOperator(_Operator):
    latency = 1

    def __init__(self, left, right):
        _Operator.__init__(self)
        self.left = _to_constant(left)
        self.right = _to_constant(right)
        self.left._add_sink(self)
        self.right._add_sink(self)
        self.op = getattr(vtypes, self.__class__.__name__, None)
        self._set_attributes()
        self._set_managers()

    def _set_attributes(self):
        left_fp = self.left.get_point()
        right_fp = self.right.get_point()
        left = self.left.bit_length() - left_fp
        right = self.right.bit_length() - right_fp
        self.width = max(left, right) + max(left_fp, right_fp)
        self.point = max(left_fp, right_fp)
        self.signed = self.left.get_signed() and self.right.get_signed()

    def _set_managers(self):
        self._set_strm(_get_strm(self.left, self.right))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def _implement(self, m, seq, svalid=None, senable=None):
        width = self.bit_length()
        signed = self.get_signed()

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        ldata, rdata = fx.adjust(self.left.sig_data, self.right.sig_data,
                                 lpoint, rpoint, signed)

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(self.op(ldata, rdata))
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(self.op(ldata, rdata)), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(self.op(ldata, rdata)), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class _UnaryOperator(_Operator):
    latency = 1

    def __init__(self, right):
        _Operator.__init__(self)
        self.right = _to_constant(right)
        self.right._add_sink(self)
        self.op = getattr(vtypes, self.__class__.__name__, None)
        self._set_attributes()
        self._set_managers()

    def _set_attributes(self):
        right = self.right.bit_length()
        right_fp = self.right.get_point()
        self.width = right
        self.point = right_fp
        self.signed = self.right.get_signed()

    def _set_managers(self):
        self._set_strm(_get_strm(self.right))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def _implement(self, m, seq, svalid=None, senable=None):
        width = self.bit_length()
        signed = self.get_signed()
        rdata = self.right.sig_data

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(self.op(rdata))
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(self.op(rdata)), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(self.op(rdata)), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class Power(_BinaryOperator):
    latency = 0

    def eval(self):
        return self.left.eval() ** self.right.eval()

    def _implement(self, m, seq, svalid=None, senable=None):
        raise NotImplementedError('_implement() is not implemented.')


class Times(_BinaryOperator):
    latency = 6 + 1

    def eval(self):
        return self.left.eval() * self.right.eval()

    def _set_attributes(self):
        left_fp = self.left.get_point()
        right_fp = self.right.get_point()
        left = self.left.bit_length()
        right = self.right.bit_length()
        self.width = max(left, right)
        self.point = min(max(left_fp, right_fp), left_fp + right_fp)
        self.signed = self.left.get_signed() and self.right.get_signed()

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency <= 3:
            raise ValueError("Latency of '*' operator must be greater than 3")

        width = self.bit_length()
        signed = self.get_signed()

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        lwidth = self.left.bit_length()
        rwidth = self.right.bit_length()
        lsigned = self.left.get_signed()
        rsigned = self.right.get_signed()
        ldata = self.left.sig_data
        rdata = self.right.sig_data

        odata = m.Wire(self.name('mul_odata'), lwidth + rwidth, signed=signed)
        odata_reg = m.Reg(self.name('mul_odata_reg'),
                          lwidth + rwidth, signed=signed, initval=0)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        shift_size = lpoint + rpoint - self.point
        if shift_size > 0:
            seq(odata_reg(fx.shift_right(odata, shift_size, signed=signed)), cond=senable)
        else:
            seq(odata_reg(odata), cond=senable)

        m.Assign(data(odata_reg))

        depth = self.latency - 1

        inst = mul.get_mul(lwidth, rwidth, lsigned, rsigned, depth)
        clk = m._clock

        update = m.Wire(self.name('mul_update'))

        if senable is not None:
            m.Assign(update(senable))
        else:
            m.Assign(update(vtypes.Int(1, 1)))

        ports = [('CLK', clk), ('update', update),
                 ('a', ldata), ('b', rdata), ('c', odata)]

        m.Instance(inst, self.name('mul'), ports=ports)


class Divide(_BinaryOperator):
    latency = 32 + 5
    variable_latency = 'get_latency'

    def get_latency(self):
        return self.bit_length() + 5

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()
        if isinstance(left, int) and isinstance(right, int):
            return int(left / right)
        return Divide(left, right)

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency <= 5:
            raise ValueError("Latency of div operator must be greater than 5")

        width = self.bit_length()
        signed = self.get_signed()

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        lsigned = self.left.get_signed()
        rsigned = self.right.get_signed()

        lval, rval = fx.adjust(self.left.sig_data, self.right.sig_data,
                               lpoint, rpoint, signed)

        ldata = m.Reg(self.name('div_ldata'), width, signed=lsigned, initval=0)
        rdata = m.Reg(self.name('div_rdata'), width, signed=rsigned, initval=0)

        seq(ldata(lval), cond=senable)
        seq(rdata(rval), cond=senable)

        sign = vtypes.Not(
            vtypes.OrList(vtypes.AndList(ldata[width - 1] == 0,
                                         rdata[width - 1] == 0),  # + , +
                          vtypes.AndList(ldata[width - 1] == 1,
                                         rdata[width - 1] == 1)))  # - , -

        abs_ldata = m.Reg(self.name('div_abs_ldata'), width, initval=0)
        abs_rdata = m.Reg(self.name('div_abs_rdata'), width, initval=0)

        if not lsigned:
            seq(abs_ldata(ldata), cond=senable)
        else:
            seq(abs_ldata(vtypes.Mux(ldata[width - 1] == 0, ldata, vtypes.Unot(ldata) + 1)),
                cond=senable)

        if not rsigned:
            seq(abs_rdata(rdata), cond=senable)
        else:
            seq(abs_rdata(vtypes.Mux(rdata[width - 1] == 0, rdata, vtypes.Unot(rdata) + 1)),
                cond=senable)

        osign = m.Wire(self.name('div_osign'))
        abs_odata = m.Wire(self.name('div_abs_odata'), width, signed=signed)
        odata = m.Reg(self.name('div_odata'), width, signed=signed, initval=0)

        if not signed:
            seq(odata(abs_odata), cond=senable)
        else:
            seq(odata(vtypes.Mux(osign == 0, abs_odata, vtypes.Unot(abs_odata) + 1)),
                cond=senable)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        m.Assign(data(odata))

        s = sign
        for i in range(self.latency - 2):
            ns = m.Reg(self.name('div_sign_tmp_%d' % i), initval=0)
            seq(ns(s), cond=senable)
            s = ns
        m.Assign(osign(s))

        inst = div.get_div()
        clk = m._clock
        rst = m._reset

        update = m.Wire(self.name('div_update'))

        if senable is not None:
            m.Assign(update(senable))
        else:
            m.Assign(update(vtypes.Int(1, 1)))

        params = [('W_D', width)]
        ports = [('CLK', clk), ('RST', rst), ('update', update), ('enable', vtypes.Int(1, 1)),
                 ('in_a', abs_ldata), ('in_b', abs_rdata), ('rslt', abs_odata)]

        m.Instance(inst, self.name('div'), params, ports)


class Mod(_BinaryOperator):
    latency = 32 + 5
    variable_latency = 'get_latency'

    def get_latency(self):
        return self.bit_length() + 5

    def eval(self):
        return self.left.eval() % self.right.eval()

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency <= 5:
            raise ValueError("Latency of div operator must be greater than 5")

        width = self.bit_length()
        signed = self.get_signed()

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        lsigned = self.left.get_signed()
        rsigned = self.right.get_signed()

        lval, rval = fx.adjust(self.left.sig_data, self.right.sig_data,
                               lpoint, rpoint, signed)

        ldata = m.Reg(self.name('mod_ldata'), width, signed=lsigned, initval=0)
        rdata = m.Reg(self.name('mod_rdata'), width, signed=rsigned, initval=0)

        seq(ldata(lval), cond=senable)
        seq(rdata(rval), cond=senable)

        sign = vtypes.Not(
            vtypes.OrList(vtypes.AndList(ldata[width - 1] == 0,
                                         rdata[width - 1] == 0),  # + , +
                          vtypes.AndList(ldata[width - 1] == 1,
                                         rdata[width - 1] == 1)))  # - , -

        abs_ldata = m.Reg(self.name('div_abs_ldata'), width, initval=0)
        abs_rdata = m.Reg(self.name('div_abs_rdata'), width, initval=0)

        if not lsigned:
            seq(abs_ldata(ldata), cond=senable)
        else:
            seq(abs_ldata(vtypes.Mux(ldata[width - 1] == 0, ldata, vtypes.Unot(ldata) + 1)),
                cond=senable)

        if not rsigned:
            seq(abs_rdata(rdata), cond=senable)
        else:
            seq(abs_rdata(vtypes.Mux(rdata[width - 1] == 0, rdata, vtypes.Unot(rdata) + 1)),
                cond=senable)

        osign = m.Wire(self.name('mod_osign'))
        abs_odata = m.Wire(self.name('mod_abs_odata'), width, signed=signed)
        odata = m.Reg(self.name('mod_odata'), width, signed=signed, initval=0)

        if not signed:
            seq(odata(abs_odata), cond=senable)
        else:
            seq(odata(vtypes.Mux(osign == 0, abs_odata, vtypes.Unot(abs_odata) + 1)),
                cond=senable)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        m.Assign(data(odata))

        s = sign
        for i in range(self.latency - 2):
            ns = m.Reg(self.name('div_sign_tmp_%d' % i), initval=0)
            seq(ns(s), cond=senable)
            s = ns
        m.Assign(osign(s))

        inst = div.get_div()
        clk = m._clock
        rst = m._reset

        update = m.Wire(self.name('mod_update'))

        if senable is not None:
            m.Assign(update(senable))
        else:
            m.Assign(update(vtypes.Int(1, 1)))

        params = [('W_D', width)]
        ports = [('CLK', clk), ('RST', rst), ('update', update), ('enable', vtypes.Int(1, 1)),
                 ('in_a', abs_ldata), ('in_b', abs_rdata), ('mod', abs_odata)]

        m.Instance(inst, self.name('div'), params, ports)


class Plus(_BinaryOperator):

    def eval(self):
        return self.left.eval() + self.right.eval()


class Minus(_BinaryOperator):

    def eval(self):
        return self.left.eval() - self.right.eval()


# general name alias
def Add(left, right):
    return Plus(left, right)


def Sub(left, right):
    return Minus(left, right)


def Mul(left, right):
    return Times(left, right)


def Div(left, right):
    return Divide(left, right)


class LessThan(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() < self.right.eval()


class GreaterThan(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() > self.right.eval()


class LessEq(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() <= self.right.eval()


class GreaterEq(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() >= self.right.eval()


class Eq(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() == self.right.eval()


class NotEq(_BinaryOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    def eval(self):
        return self.left.eval() != self.right.eval()


class _BinaryShiftOperator(_BinaryOperator):

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.right.get_point() != 0:
            raise TypeError("shift amount must be int")

        width = self.bit_length()
        signed = self.get_signed()

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        ldata, rdata = self.left.sig_data, self.right.sig_data

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(self.op(ldata, rdata))
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(self.op(ldata, rdata)), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(self.op(ldata, rdata)), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class Sll(_BinaryShiftOperator):
    max_width = 1024

    def _set_attributes(self):
        v = self.right.eval()
        if isinstance(v, int):
            width = self.left.bit_length() + v
        else:
            v = 2 ** self.right.bit_length()
            width = self.left.bit_length() + v

        if width > self.max_width:
            raise ValueError("bit_length is too large '%d'" % width)

        self.width = width
        self.point = self.left.get_point()
        self.signed = self.left.get_signed()

    def eval(self):
        return self.left.eval() << self.right.eval()


class Srl(_BinaryShiftOperator):

    def _set_attributes(self):
        self.width = self.left.bit_length()
        self.point = self.left.get_point()
        self.signed = False

    def eval(self):
        return self.left.eval() >> self.right.eval()


class Sra(_BinaryShiftOperator):

    def _set_attributes(self):
        self.width = self.left.bit_length()
        self.point = self.left.get_point()
        self.signed = self.left.get_signed()

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()
        if isinstance(left, int) and isinstance(right, int):
            sign = left >= 0
            left = abs(left)
            ret = left >> right
            if not sign:
                return -1 * ret
            return ret
        return Sra(left, right)


class _BinaryLogicalOperator(_BinaryOperator):

    def _set_attributes(self):
        left = self.left.bit_length()
        right = self.right.bit_length()
        self.width = max(left, right)
        self.point = 0
        self.signed = False

    def _implement(self, m, seq, svalid=None, senable=None):
        width = self.bit_length()
        signed = False

        lpoint = self.left.get_point()
        rpoint = self.right.get_point()
        ldata, rdata = self.left.sig_data, self.right.sig_data

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(self.op(ldata, rdata))
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(self.op(ldata, rdata)), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(self.op(ldata, rdata)), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class And(_BinaryLogicalOperator):

    def eval(self):
        return self.left.eval() & self.right.eval()


class Xor(_BinaryLogicalOperator):

    def eval(self):
        return self.left.eval() ^ self.right.eval()


class Xnor(_BinaryLogicalOperator):

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()
        ret = left ^ right
        if isinstance(ret, int):
            return ret == 0
        return Xnor(left, right)


class Or(_BinaryLogicalOperator):

    def eval(self):
        return self.left.eval() | self.right.eval()


class Land(_BinaryLogicalOperator):

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()
        if isinstance(left, (int, bool)) and isinstance(right, (int, bool)):
            return left and right
        return Land(left, right)


class Lor(_BinaryLogicalOperator):

    def eval(self):
        left = self.left.eval()
        right = self.right.eval()
        if isinstance(left, (int, bool)) and isinstance(right, (int, bool)):
            return left or right
        return Land(left, right)


class Uplus(_UnaryOperator):

    def eval(self):
        return self.right.eval()


class Uminus(_UnaryOperator):

    def eval(self):
        return - self.right.eval()


class _UnaryLogicalOperator(_UnaryOperator):

    def _set_attributes(self):
        right = self.right.bit_length()
        self.width = right
        self.point = 0
        self.signed = False


class Ulnot(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, (int, bool)):
            return not right
        return Ulnot(right)


class Unot(_UnaryLogicalOperator):

    def eval(self):
        right = self.right.eval()
        try:
            v = ~right
        except:
            v = Ulnot(right)
        return v


class Uand(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return right
        if isinstance(right, int):
            width = self.right.bit_length()
            for i in range(width):
                if right & 0x1 == 0:
                    return False
                right = right >> 1
            return True
        return Uand(right)


class Unand(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return not right
        if isinstance(right, int):
            width = self.right.bit_length()
            for i in range(width):
                if right & 0x1 == 0:
                    return True
                right = right >> 1
            return False
        return Unand(right)


class Uor(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return right
        if isinstance(right, int):
            width = self.right.bit_length()
            for i in range(width):
                if right & 0x1 == 1:
                    return True
                right = right >> 1
            return False
        return Uor(right)


class Unor(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return not right
        if isinstance(right, int):
            width = self.right.bit_length()
            for i in range(width):
                if right & 0x1 == 1:
                    return False
                right = right >> 1
            return True
        return Unor(right)


class Uxor(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return right
        if isinstance(right, int):
            width = self.right.bit_length()
            ret = 1
            for i in range(width):
                ret = ret ^ (right & 0x1)
                right = right >> 1
            return ret == 1
        return Uxor(right)


class Uxnor(_UnaryLogicalOperator):

    def _set_attributes(self):
        self.width = 1
        self.point = 0

    def eval(self):
        right = self.right.eval()
        if isinstance(right, bool):
            return not right
        if isinstance(right, int):
            width = self.right.bit_length()
            ret = 1
            for i in range(width):
                ret = ret ^ (right & 0x1)
                right = right >> 1
            return ret == 0
        return Uxnor(right)


# alias
def Not(*args):
    return Ulnot(*args)


def AndList(*args):
    if len(args) == 0:
        raise ValueError("LandList requires at least one argument.")
    if len(args) == 1:
        return args[0]
    left = args[0]
    for right in args[1:]:
        left = Land(left, right)
    return left


def OrList(*args):
    if len(args) == 0:
        raise ValueError("LorList requires at least one argument.")
    if len(args) == 1:
        return args[0]
    left = args[0]
    for right in args[1:]:
        left = Lor(left, right)
    return left


Ands = AndList
Ors = OrList


class Cast(_UnaryOperator):
    latency = 0

    def __init__(self, data, width=None, point=None, signed=None):
        _UnaryOperator.__init__(self, data)
        if width is not None:
            self.width = width
        if point is not None:
            self.point = point
        if signed is not None:
            self.signed = signed

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 0:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 0))

        width = self.bit_length()
        signed = self.get_signed()

        rdata = self.right.sig_data
        rwidth = self.right.bit_length()
        rpoint = self.right.get_point()
        rsigned = self.right.get_signed()

        rdata_src = m.Wire(self.name('src'), rwidth, signed=rsigned)
        rdata_src.assign(rdata)

        if rpoint > self.point:
            rdata_src = fx.shift_right(rdata_src, rpoint - self.point, rsigned)
        elif rpoint < self.point:
            rdata_src = fx.shift_left(rdata_src, self.point - rpoint, rsigned)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        m.Assign(data(rdata_src))

    def eval(self):
        return self


class ReinterpretCast(Cast):

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 0:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 0))

        width = self.bit_length()
        signed = self.get_signed()

        rdata = self.right.sig_data
        rwidth = self.right.bit_length()
        rsigned = self.right.get_signed()

        rdata_src = m.Wire(self.name('src'), rwidth, signed=rsigned)
        rdata_src.assign(rdata)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        m.Assign(data(rdata_src))

    def eval(self):
        return self


class _SpecialOperator(_Operator):
    latency = 1

    def __init__(self, *args):
        _Operator.__init__(self)
        self.args = [_to_constant(arg) for arg in args]
        for var in self.args:
            var._add_sink(self)
        self.op = None
        self._set_attributes()
        self._set_managers()

        self.graph_shape = 'ellipse'

    def _set_attributes(self):
        if len(self.args) > 1:
            wargs = [arg.bit_length() for arg in self.args]
            self.width = max(*wargs)
            pargs = [arg.get_point() for arg in self.args]
            self.point = max(*pargs)
        elif self.args:
            self.width = self.args[0].bit_length()
            self.point = self.args[0].get_point()
        else:
            self.width = 1
            self.point = 0

        self.signed = False
        for arg in self.args:
            if arg.get_signed():
                self.signed = True
                break

    def _set_managers(self):
        self._set_strm(_get_strm(*self.args))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def _implement(self, m, seq, svalid=None, senable=None):
        width = self.bit_length()
        signed = self.get_signed()

        arg_data = [arg.sig_data for arg in self.args]

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(self.op(*arg_data))
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(self.op(*arg_data)), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(self.op(*arg_data)), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class Pointer(_SpecialOperator):

    def __init__(self, var, pos):
        _SpecialOperator.__init__(self, var, pos)
        self.op = vtypes.Pointer

    def _set_attributes(self):
        self.width = 1
        self.point = 0
        self.signed = False

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    @property
    def pos(self):
        return self.args[1]

    @pos.setter
    def pos(self, pos):
        self.args[1] = pos

    def eval(self):
        var = self.var.eval()
        pos = self.pos.eval()
        if isinstance(var, int) and isinstance(pos, int):
            return (var >> pos) & 0x1
        return Pointer(var, pos)


class Slice(_SpecialOperator):

    def __init__(self, var, msb, lsb):
        msb = msb.eval() if isinstance(msb, _Constant) else msb
        lsb = lsb.eval() if isinstance(lsb, _Constant) else lsb
        if not isinstance(msb, int) or not isinstance(lsb, int):
            raise TypeError('msb and lsb must be int')
        msb = Int(msb, signed=False)
        lsb = Int(lsb, signed=False)
        _SpecialOperator.__init__(self, var, msb, lsb)
        self.op = vtypes.Slice

    def _set_attributes(self):
        self.width = self.msb.eval() - self.lsb.eval() + 1
        self.point = 0
        self.signed = False

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    @property
    def msb(self):
        return self.args[1]

    @msb.setter
    def msb(self, msb):
        self.args[1] = msb

    @property
    def lsb(self):
        return self.args[2]

    @lsb.setter
    def lsb(self, lsb):
        self.args[2] = lsb

    def eval(self):
        var = self.var.eval()
        msb = self.msb.eval()
        lsb = self.lsb.eval()
        if isinstance(var, int) and isinstance(msb, int) and isinstance(lsb, int):
            mask = 0
            for i in range(msb - lsb + 1):
                mask = (mask << 1) | 0x1
            return (var >> lsb) & mask
        return Slice(var, msb, lsb)


def Split(data, width=None, point=None, signed=None, num_chunks=None, reverse=False):
    """
    Split the given data into multiple chunks

    Parameters
    ----------
    data : _Numeric
        Input data

    width : int
        Data width of separated chunks (default: same as input data)

    point : int
        Fixed-point position of separated chunks (default: same as input data)

    signed : bool
        Sign (default: same as input data)

    num_chunks: int
        The number of separated chunks (default: (input data width) / width)

    reverse: bool
        reverse flag. If true, a reversed list is returned

    Returns
    -------
    chunks : list
        A list of separated chunks

        For the consistency with Cat operator, the order of chunks is higher-bit first.
        If data is a 32-bit value, and width is 8, returned list of chunks will be
            chunks = [data[31:24], data[23:16], data[15:8], data[7:0]]
        If reverse == True, it returns the reversed list of chunks
    """

    data = _to_constant(data)

    if width is None and num_chunks is None:
        raise ValueError('width or num_chunks must be specified.')

    if width is not None and num_chunks is not None:
        raise ValueError('Either of width or num_chunks must be specified.')

    if width is None:
        width = int(ceil(data.bit_length() / num_chunks))
    elif num_chunks is None:
        num_chunks = int(ceil(data.bit_length() / width))

    if point is None:
        point = data.get_point()

    if signed is None:
        signed = data.get_signed()

    total_width = width * num_chunks
    ret = []
    for i in range(0, total_width, width):
        if i + width > data.bit_length():
            if signed:
                sign = data[-1]
                sign.latency = 0
                pad = Repeat(sign, total_width - data.bit_length())
                pad.latency = 0
            else:
                pad = Int(0, signed=False)
                pad.width = total_width - data.bit_length()

            slc = Slice(data, data.bit_length() - 1, i)
            slc.latency = 0
            v = Cat(pad, slc)
            v.latency = 0
        else:
            v = Slice(data, i + width - 1, i)
            v.latency = 0

        ret.append(ReinterpretCast(v, width, point, signed))

    if not reverse:
        ret.reverse()

    return ret


class Cat(_SpecialOperator):

    def __init__(self, *vars):
        _SpecialOperator.__init__(self, *vars)
        self.op = vtypes.Cat

    def _set_attributes(self):
        ret = 0
        for v in self.vars:
            ret += v.bit_length()
        self.width = ret
        self.point = 0
        self.signed = False

    @property
    def vars(self):
        return self.args

    @vars.setter
    def vars(self, vars):
        self.args = list(vars)

    def eval(self):
        vars = [var.eval() for var in self.vars]
        for var in vars:
            if not isinstance(var, int):
                return Cat(*vars)
        ret = 0
        for var in vars:
            ret = (ret << var.bit_length()) | var
        return ret


class Repeat(_SpecialOperator):

    def __init__(self, var, times):
        times = times.eval() if isinstance(times, _Constant) else times
        if not isinstance(times, int):
            raise TypeError('times must be int')
        _SpecialOperator.__init__(self, var, times)
        self.op = vtypes.Repeat

    def _set_attributes(self):
        self.width = self.var.bit_length() * self.times.eval()
        self.point = 0
        self.signed = False

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    @property
    def times(self):
        return self.args[1]

    @times.setter
    def times(self, times):
        self.args[1] = times

    def eval(self):
        var = self.var.eval()
        times = self.times.eval()
        ret = 0
        for i in times:
            ret = (ret << var.bit_length()) | var
        return ret


class Cond(_SpecialOperator):

    def __init__(self, condition, true_value, false_value):
        _SpecialOperator.__init__(self, condition, true_value, false_value)
        self.op = vtypes.Cond

    def _set_attributes(self):
        true_value_fp = self.true_value.get_point()
        false_value_fp = self.false_value.get_point()
        true_value = self.true_value.bit_length() - true_value_fp
        false_value = self.false_value.bit_length() - false_value_fp
        self.width = max(true_value, false_value) + \
            max(true_value_fp, false_value_fp)
        self.point = max(true_value_fp, false_value_fp)
        self.signed = self.true_value.get_signed() or self.false_value.get_signed()

    @property
    def condition(self):
        return self.args[0]

    @condition.setter
    def condition(self, condition):
        self.args[0] = condition

    @property
    def true_value(self):
        return self.args[1]

    @true_value.setter
    def true_value(self, true_value):
        self.args[1] = true_value

    @property
    def false_value(self):
        return self.args[2]

    @false_value.setter
    def false_value(self, false_value):
        self.args[2] = false_value

    def eval(self):
        condition = self.condition.eval()
        true_value = self.true_value.eval()
        false_value = self.false_value.eval()
        if isinstance(condition, (int, bool)):
            if condition:
                return true_value
            else:
                return false_value
        return Cond(condition, true_value, false_value)


def Mux(condition, true_value, false_value):
    # return the result immediately if the condition can be resolved now
    if isinstance(condition, (bool, int, float, str, list, tuple)):
        return true_value if condition else false_value
    return Cond(condition, true_value, false_value)


class CustomOp(_SpecialOperator):

    def __init__(self, op, *vars):
        _SpecialOperator.__init__(self, *vars)
        self.op = op

    def eval(self):
        return self


class LUT(_SpecialOperator):
    latency = 1

    def __init__(self, address, patterns, width=32, point=0, signed=True):
        _SpecialOperator.__init__(self, address)
        self.op = None
        self.width = width
        self.point = point
        self.signed = signed
        self.patterns = patterns
        self.graph_label = 'LUT'

    def _set_attributes(self):
        pass

    @property
    def address(self):
        return self.args[0]

    @address.setter
    def address(self, address):
        self.args[0] = address

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        width = self.bit_length()
        signed = self.get_signed()

        arg_data = self.address.sig_data

        size = int(log(len(self.patterns), 2))
        address = m.Wire(self.name('lut_address'), width=size)
        address.assign(arg_data)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        inst = rom.mkROMDefinition(self.name('LUT_ROM'), self.patterns,
                                   size, width, sync=True, with_enable=True)
        clk = m._clock

        ports = [('CLK', clk), ('addr', address),
                 ('enable', senable), ('val', data)]

        m.Instance(inst, self.name('lut'), ports=ports)


class Complement2(_SpecialOperator):

    def __init__(self, var):
        _SpecialOperator.__init__(self, var)
        self.op = vtypes.Complement2

    def _set_attributes(self):
        self.width = self.var.bit_length()
        self.point = self.var.get_point()
        self.signed = self.var.get_signed()

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    def eval(self):
        var = self.var.eval()
        ret = Complement2(var)
        return ret


class Abs(_SpecialOperator):

    def __init__(self, var):
        _SpecialOperator.__init__(self, var)
        self.op = vtypes.Abs

    def _set_attributes(self):
        self.width = self.var.bit_length()
        self.point = self.var.get_point()
        self.signed = self.var.get_signed()

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    def eval(self):
        var = self.var.eval()
        ret = abs(var)
        return ret


class Sign(_SpecialOperator):

    def __init__(self, var):
        _SpecialOperator.__init__(self, var)
        self.op = vtypes.Sign

    def _set_attributes(self):
        self.width = self.var.bit_length()
        self.point = self.var.get_point()
        self.signed = self.var.get_signed()

    @property
    def var(self):
        return self.args[0]

    @var.setter
    def var(self, var):
        self.args[0] = var

    def eval(self):
        var = self.var.eval()
        ret = Sign(var)
        return ret


class _Delay(_UnaryOperator):
    latency = 1

    def __init__(self, right):
        _UnaryOperator.__init__(self, right)
        # parent value for delayed_value and previous_value
        self.parent_value = None

        self.graph_label = 'Delay'
        self.graph_shape = 'box'
        self.graph_color = 'lightgray'
        self.graph_style = 'filled'

    def _set_parent_value(self, value):
        self.parent_value = value

    def _get_parent_value(self):
        return self.parent_value

    def eval(self):
        return self

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        width = self.bit_length()
        signed = self.get_signed()
        rdata = self.right.sig_data

        data = m.Reg(self.name('data'), width, initval=0, signed=signed)
        self.sig_data = data

        seq(data(rdata), cond=senable)


class _Prev(_UnaryOperator):
    latency = 0

    def __init__(self, right):
        _UnaryOperator.__init__(self, right)
        # parent value for delayed_value and previous_value
        self.parent_value = None

        self.graph_label = 'Prev'
        self.graph_shape = 'box'

    def _set_parent_value(self, value):
        self.parent_value = value

    def _get_parent_value(self):
        return self.parent_value

    def eval(self):
        return self

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 0:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 0))

        width = self.bit_length()
        signed = self.get_signed()
        rdata = self.right.sig_data

        data = m.Reg(self.name('data'), width, initval=0, signed=signed)
        self.sig_data = data

        seq(data(rdata), cond=senable)


class _PlusN(_SpecialOperator):
    latency = 1

    def __init__(self, *vars):
        _SpecialOperator.__init__(self, *vars)

        for arg in self.args:
            if arg.point != 0:
                raise ValueError('Fixed point is not supported.')

        def plus_n(*args):
            ret = args[0]
            for arg in args[1:]:
                ret += arg
            return ret

        self.op = plus_n
        self.graph_label = 'PlusN'

    def eval(self):
        vars = [var.eval() for var in self.vars]
        for var in vars:
            if not isinstance(var, int):
                return PlusN(*vars)
        ret = 0
        for var in vars:
            ret += var
        return ret


class _MulAdd(_SpecialOperator):
    latency = 6 + 1

    def __init__(self, a, b, c):
        _SpecialOperator.__init__(self, a, b, c)

        if self.a.point + self.b.point != self.c.point:
            raise ValueError('Unsupported fixed point combination')

        self.graph_label = 'MulAdd'

    @property
    def a(self):
        return self.args[0]

    @a.setter
    def a(self, a):
        self.args[0] = a

    @property
    def b(self):
        return self.args[1]

    @b.setter
    def b(self, b):
        self.args[1] = b

    @property
    def c(self):
        return self.args[2]

    @c.setter
    def c(self, c):
        self.args[2] = c

    def _set_attributes(self):
        a_fp = self.a.get_point()
        b_fp = self.b.get_point()
        c_fp = self.c.get_point()
        a = self.a.bit_length()
        b = self.b.bit_length()
        c = self.c.bit_length()
        self.width = max(a, b, c)
        self.point = a_fp + b_fp
        self.signed = self.a.get_signed() and self.b.get_signed() and self.c.get_signed()

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency <= 3:
            raise ValueError("Latency of '*' operator must be greater than 3")

        width = self.bit_length()
        signed = self.get_signed()

        apoint = self.a.get_point()
        bpoint = self.b.get_point()
        cpoint = self.c.get_point()
        awidth = self.a.bit_length()
        bwidth = self.b.bit_length()
        cwidth = self.c.bit_length()
        asigned = self.a.get_signed()
        bsigned = self.b.get_signed()
        csigned = self.c.get_signed()
        adata = self.a.sig_data
        bdata = self.b.sig_data
        cdata = self.c.sig_data

        odata = m.Wire(self.name('madd_odata'),
                       max(awidth + bwidth, cwidth), signed=signed)
        odata_reg = m.Reg(self.name('madd_odata_reg'),
                          max(awidth + bwidth, cwidth), signed=signed, initval=0)

        data = m.Wire(self.name('data'), width, signed=signed)
        self.sig_data = data

        seq(odata_reg(odata), cond=senable)

        m.Assign(data(odata_reg))

        depth = self.latency - 1

        inst = madd.get_madd(awidth, bwidth, cwidth,
                             asigned, bsigned, csigned, depth)
        clk = m._clock

        update = m.Wire(self.name('madd_update'))

        if senable is not None:
            m.Assign(update(senable))
        else:
            m.Assign(update(vtypes.Int(1, 1)))

        ports = [('CLK', clk), ('update', update),
                 ('a', adata), ('b', bdata), ('c', cdata), ('d', odata)]

        m.Instance(inst, self.name('madd'), ports=ports)

    def eval(self):
        vars = [var.eval() for var in self.vars]
        for var in vars:
            if not isinstance(var, int):
                return MulAdd(*vars)

        return vars[0] * vars[1] + vars[2]


def MulAdd(a, b, c):
    a_point = a.point if isinstance(a, _Numeric) else 0
    b_point = b.point if isinstance(b, _Numeric) else 0
    c_point = c.point if isinstance(c, _Numeric) else 0
    if a_point + b_point != c_point:
        return Plus(Times(a, b), c)

    return _MulAdd(a, b, c)


def Madd(a, b, c):
    return MulAdd(a, b, c)


def op_tree(op, initval, latency, *args):
    if len(args) == 0:
        return initval

    if len(args) == 1:
        return args[0]

    half_len = len(args) // 2
    ret = op(op_tree(op, initval, latency, *args[:half_len]),
             op_tree(op, initval, latency, *args[half_len:]))

    if latency is not None:
        ret.latency = latency

    return ret


def PlusN(*args):
    for arg in args:
        if isinstance(arg, _Numeric) and arg.point != 0:
            ret = op_tree(Plus, Int(0, signed=True), 0, *args)
            ret.latency = 1
            return ret

    return _PlusN(*args)


def AddN(*args):
    return PlusN(*args)


def AddTree(*args):
    return op_tree(Plus, Int(0, signed=True), None, *args)


def Max(*args):
    if len(args) == 1:
        return args[0]
    initval = args[0]
    return op_tree(lambda x, y: Mux(x > y, x, y), initval, None, *args)


def Min(*args):
    if len(args) == 1:
        return args[0]
    initval = args[0]
    return op_tree(lambda x, y: Mux(x < y, x, y), initval, None, *args)


def Average(*args):
    sum = AddTree(*args)
    length = len(args)
    if length & (length - 1) == 0:
        return Sra(sum, int(log(length, 2)))
    return Div(sum, Int(length, signed=True))


def AverageRound(*args):
    sum = AddTree(*args)
    length = len(args)
    sum += length // 2
    if length & (length - 1) == 0:
        return Sra(sum, int(log(length, 2)))
    return Div(sum, Int(length, signed=True))


def SraRound(left, right):
    shifted = Sra(left, right)
    last_bit = Sra(left, right - 1)[0]
    return Mux(right == Int(0), left, shifted + last_bit)


class _Constant(_Numeric):

    def __init__(self, value):
        _Numeric.__init__(self)
        self.value = value
        self.signed = False
        self._set_attributes()
        self._set_managers()
        self.sig_data = self.value

        self.graph_shape = 'box'
        self.graph_color = 'lightblue'
        self.graph_style = 'rounded,filled'
        self.graph_peripheries = 1

    def _set_attributes(self):
        self.width = self.value.bit_length() + 1
        self.point = 0
        self.signed = False

    def _set_managers(self):
        self._set_strm(_get_strm(self.value))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def eval(self):
        return self.value

    def _implement(self, m, seq, svalid=None, senable=None):
        data = self.value
        self.sig_data = data


class _Variable(_Numeric):

    def __init__(self, data=None, width=32, point=0, signed=True):
        _Numeric.__init__(self)
        self.input_data = data
        if isinstance(self.input_data, _Numeric):
            self.input_data._add_sink(self)
        self.width = width
        self.point = point
        self.signed = signed

        if isinstance(self.input_data, _Numeric):
            self.graph_label = self.input_data.graph_label
        else:
            inobj = str(self.input_data)
            label_data = [inobj, str(self.width)]
            if self.point > 0:
                label_data.append(str(self.point))
            self.graph_label = ':'.join(label_data)

        self.graph_shape = 'box'
        self.graph_color = 'lightblue'
        self.graph_style = 'filled'
        self.graph_peripheries = 2

    def eval(self):
        return self

    def output(self, data):
        if isinstance(self.input_data, _Numeric):
            self.input_data.output(data)
            return
        _Numeric.output(self, data)

    def connect(self, data):
        if self.sig_data is not None:
            raise ValueError("Input signals are already synthesized.")

        if not isinstance(data, (_Numeric, vtypes._Numeric, int, bool)):
            raise TypeError(
                "'data' must be stypes._Numeric or vtypes._Numeric.")

        self.input_data = data
        if isinstance(self.input_data, _Numeric):
            self.input_data._add_sink(self)

    def write(self, wdata, cond=None):
        if self.sig_data is None:
            if self.m is None:
                raise ValueError("Module information is not set.")
            self._implement_input(self.m, self.seq, aswire=True)

        if isinstance(self.sig_data, vtypes.Input):
            raise TypeError("Variable with Input type is not supported.")

        if isinstance(self.sig_data, vtypes.Wire):
            width = self.bit_length()
            signed = self.get_signed()
            if hasattr(self, 'sig_data_write'):
                data = self.sig_data_write
            else:
                data = self.m.Reg(self.name('wdata'),
                                  width, initval=0, signed=signed)
                self.sig_data_write = data
                self.sig_data.assign(data)
        else:
            data = self.sig_data

        if cond is not None:
            self.seq.If(cond)

        self.seq(
            data(wdata)
        )

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.input_data is None:
            raise TypeError("'input_data' must not be None.")

        # if input_data is a standard signal, skip
        if not isinstance(self.input_data, _Numeric):
            return

        if self.input_data.sig_data is None:
            self.input_data._implement(m, seq)

        self.sig_data = self.input_data.sig_data

    def _implement_input(self, m, seq, aswire=False):
        if self.input_data is None:
            raise TypeError("'input_data' must not be None")

        # if input_data is an other variable, skip
        if isinstance(self.input_data, _Numeric):
            return

        # if already synthesized
        if self.sig_data is not None:
            return

        type_i = m.Wire if aswire else m.Input
        type_o = m.Wire if aswire else m.Output

        width = self.bit_length()
        signed = self.get_signed()

        if isinstance(self.input_data, (vtypes._Numeric, int, bool)):
            self.sig_data = self.input_data
        else:
            self.sig_data = type_i(self.input_data, width, signed=signed)

    def _implement_output(self, m, seq, aswire=False):
        if isinstance(self.input_data, _Numeric):
            if self.input_data.output_sig_data is None:
                self.input_data._implement_output(m, seq, aswire)
            self.output_sig_data = self.input_data.output_sig_data
            return
        _Numeric._implement_output(self, m, seq, aswire)

    # __getattribute__() method is always called,
    # whenever fields of the node is accessed.
    def __getattribute__(self, attr):
        # for isinstance method
        if attr == '__class__':
            return _Numeric.__getattribute__(self, '__class__')

        try:
            input_data = _Numeric.__getattribute__(self, 'input_data')
        except AttributeError:
            return _Numeric.__getattribute__(self, attr)

        # always returns input_data for 'input_data' attribute
        if attr == 'input_data':
            return input_data

        # if it has a variable alias, redirect to it
        if isinstance(input_data, _Numeric):
            return getattr(input_data, attr)

        # nornal access
        return _Numeric.__getattribute__(self, attr)


class _ParameterVariable(_Variable):

    def __init__(self, data, width=32, point=0, signed=True, value=None):
        if isinstance(data, _Numeric):
            raise TypeError(
                "_ParameterVariable cannot receive type '%s'" % str(type(data)))

        if value is not None and not isinstance(data, str):
            raise TypeError(
                "Required str for 'data', when 'value' is assigned")

        _Variable.__init__(self, data=data, width=width,
                           point=point, signed=signed)
        self.value = value

        inobj = str(self.input_data)
        label_data = [inobj, str(self.width)]
        if self.point > 0:
            label_data.append(str(self.point))
        self.graph_label = ':'.join(label_data)

        self.graph_shape = 'circle'
        self.graph_color = 'lightblue'
        self.graph_style = 'filled'
        self.graph_peripheries = 1

    def _implement(self, m, seq, svalid=None, senable=None):
        pass

    def _implement_input(self, m, seq, aswire=False):
        type_i = m.Wire if aswire else m.Input

        width = self.bit_length()
        signed = self.get_signed()

        if isinstance(self.input_data, (vtypes._Numeric, int, bool)):
            self.sig_data = self.input_data
        elif self.value is not None:
            self.sig_data = m.Parameter(self.input_data, self.value,
                                        width=self.width, signed=self.signed)
        else:
            self.sig_data = type_i(
                self.input_data, self.width, signed=self.signed)

    def __getattribute__(self, attr):
        # normal access
        return _Numeric.__getattribute__(self, attr)


class _Accumulator(_UnaryOperator):
    latency = 1
    ops = (vtypes.Plus, )

    def __init__(self, right, size=None, initval=None,
                 enable=None, reset=None, width=32, signed=True):

        self.size = _to_constant(size) if size is not None else None
        self.initval = (_to_constant(initval)
                        if initval is not None else _to_constant(0))

        if not isinstance(self.initval, _Constant):
            raise TypeError("initval must be Constant, not '%s'" %
                            str(type(self.initval)))

        self.enable = _to_constant(enable)
        if self.enable is not None:
            self.enable._add_sink(self)

        self.reset = _to_constant(reset)
        if self.reset is not None:
            self.reset._add_sink(self)

        _UnaryOperator.__init__(self, right)
        self.width = width
        self.signed = signed

        self.graph_shape = 'box'
        self.graph_style = 'rounded'

    def _set_attributes(self):
        self.point = self.right.get_point()

    def _set_managers(self):
        self._set_strm(_get_strm(self.right, self.initval,
                                 self.enable, self.reset))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def eval(self):
        return self

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        size_data = self.size.sig_data if self.size is not None else None
        initval_data = self.initval.sig_data

        width = self.bit_length()
        signed = self.get_signed()

        # for Pulse
        if not self.ops and self.size is not None:
            width = 1

        data = m.Reg(self.name('data'), width,
                     initval=initval_data, signed=signed)

        if self.size is not None:
            count = m.Reg(self.name('count'),
                          size_data.bit_length() + 1, initval=0)
            next_count_value = vtypes.Mux(count >= size_data - 1,
                                          0, count + 1)
            count_zero = (count == 0)

        self.sig_data = data

        rdata = self.right.sig_data
        enabledata = self.enable.sig_data if self.enable is not None else None
        resetdata = self.reset.sig_data if self.reset is not None else None

        value = data
        for op in self.ops:
            if not isinstance(op, type):
                value = op(value, rdata)
            elif issubclass(op, vtypes._BinaryOperator):
                value = op(value, rdata)
            elif issubclass(op, vtypes._UnaryOperator):
                value = op(value)

            if not isinstance(value, vtypes._Numeric):
                raise TypeError("Operator '%s' returns unsupported object type '%s'."
                                % (str(op), str(type(value))))

        # for Pulse
        if not self.ops and self.size is not None:
            value = (count >= (size_data - 1))

        if self.reset is not None or self.size is not None:
            reset_value = initval_data
            for op in self.ops:
                if not isinstance(op, type):
                    reset_value = op(reset_value, rdata)
                elif issubclass(op, vtypes._BinaryOperator):
                    reset_value = op(reset_value, rdata)
                elif issubclass(op, vtypes._UnaryOperator):
                    reset_value = op(reset_value)

                if not isinstance(reset_value, vtypes._Numeric):
                    raise TypeError("Operator '%s' returns unsupported object type '%s'."
                                    % (str(op), str(type(reset_value))))

            if not self.ops and self.size is not None:
                reset_value = (count >= (size_data - 1))

        if self.enable is not None:
            enable_cond = _and_vars(svalid, senable, enabledata)
            seq(data(value), cond=enable_cond)

            if self.size is not None:
                seq(count(next_count_value), cond=enable_cond)

        else:
            enable_cond = _and_vars(svalid, senable)
            seq(data(value), cond=enable_cond)

            if self.size is not None:
                seq(count(next_count_value), cond=enable_cond)

        if self.reset is not None:
            if self.enable is None:
                reset_cond = _and_vars(svalid, senable, resetdata)
                seq(data(reset_value), cond=reset_cond)

                if self.size is not None:
                    seq(count(0), cond=reset_cond)
                    reset_cond = _and_vars(svalid, senable, count_zero)
                    seq(data(reset_value), cond=reset_cond)

            else:
                reset_cond = _and_vars(svalid, senable, resetdata)
                seq(data(initval_data), cond=reset_cond)

                reset_enable_cond = _and_vars(
                    svalid, senable, enabledata, resetdata)
                seq(data(reset_value), cond=reset_enable_cond)

                if self.size is not None:
                    seq(count(0), cond=reset_enable_cond)
                    reset_enable_cond = _and_vars(
                        svalid, senable, enabledata, count_zero)
                    seq(data(reset_value), cond=reset_enable_cond)

        elif self.size is not None:
            if self.enable is not None:
                reset_enable_cond = _and_vars(
                    svalid, senable, enabledata, count_zero)
                seq(data(reset_value), cond=reset_enable_cond)
            else:
                reset_cond = _and_vars(svalid, senable, count_zero)
                seq(data(reset_value), cond=reset_cond)


class ReduceAdd(_Accumulator):
    ops = (vtypes.Plus, )

    def __init__(self, right, size=None, initval=0,
                 enable=None, reset=None, width=32, signed=True):
        _Accumulator.__init__(self, right, size, initval,
                              enable, reset, width, signed)


class ReduceSub(_Accumulator):
    ops = (vtypes.Minus, )

    def __init__(self, right, size=None, initval=0,
                 enable=None, reset=None, width=32, signed=True):
        _Accumulator.__init__(self, right, size, initval,
                              enable, reset, width, signed)


class ReduceMul(_Accumulator):
    latency = 1
    ops = (vtypes.Times, )

    def __init__(self, right, size=None, initval=0,
                 enable=None, reset=None, width=32, signed=True):
        _Accumulator.__init__(self, right, size, initval,
                              enable, reset, width, signed)


class ReduceDiv(_Accumulator):
    latency = 32
    ops = ()

    def __init__(self, right, size=None, initval=0,
                 enable=None, reset=None, width=32, signed=True):
        raise NotImplementedError()
        _Accumulator.__init__(self, right, size, initval,
                              enable, reset, width, signed)


class ReduceCustom(_Accumulator):

    def __init__(self, ops, right, size=None, initval=0,
                 enable=None, reset=None, width=32, signed=True, label=None):
        _Accumulator.__init__(self, right, size, initval,
                              enable, reset, width, signed)
        if not isinstance(ops, (tuple, list)):
            ops = tuple([ops])
        self.ops = ops
        self.graph_label = label


class Counter(_Accumulator):

    def __init__(self, size=None, step=1, initval=0,
                 control=None, enable=None, reset=None, width=32, signed=False):

        self.ops = (lambda x, y: x + step, )

        if control is None:
            control = 0

        initval -= step

        _Accumulator.__init__(self, control, size, initval,
                              enable, reset, width, signed)
        self.graph_label = 'Counter'


class Pulse(_Accumulator):
    ops = ()

    def __init__(self, size, control=None, enable=None, reset=None):

        if control is None:
            control = 0

        step = 1
        initval = 0
        width = 1
        signed = False

        _Accumulator.__init__(self, control, size, initval,
                              enable, reset, width, signed)
        self.graph_label = 'Pulse'


def _ReduceValid(cls, right, size, initval=0,
                 enable=None, reset=None, width=32, signed=True):

    data = cls(right, size, initval,
               enable, reset, width, signed)
    valid = Pulse(size, right, enable, reset)

    return data, valid


def ReduceAddValid(right, size, initval=0,
                   enable=None, reset=None, width=32, signed=True):

    cls = ReduceAdd
    return _ReduceValid(cls, right, size, initval,
                        enable, reset, width, signed)


def ReduceSubValid(right, size, initval=0,
                   enable=None, reset=None, width=32, signed=True):

    cls = ReduceSub
    return _ReduceValid(cls, right, size, initval,
                        enable, reset, width, signed)


def ReduceMulValid(right, size, initval=0,
                   enable=None, reset=None, width=32, signed=True):

    cls = ReduceMul
    return _ReduceValid(cls, right, size, initval,
                        enable, reset, width, signed)


def ReduceDivValid(right, size, initval=0,
                   enable=None, reset=None, width=32, signed=True):

    cls = ReduceDiv
    return _ReduceValid(cls, right, size, initval,
                        enable, reset, width, signed)


def ReduceCustomValid(ops, right, size, initval=0,
                      enable=None, reset=None, width=32, signed=True):

    data = ReduceCustom(ops, right, size, initval,
                        enable, reset, width, signed)
    valid = Pulse(size, right, enable, reset)

    return data, valid


def CounterValid(size, step=1, initval=0,
                 control=None, enable=None, reset=None, width=32, signed=False):

    data = Counter(size, step, initval,
                   control, enable, reset, width, signed)
    valid = Pulse(size, control, enable, reset)

    return data, valid


class Int(_Constant):

    def __init__(self, value, signed=True):
        _Constant.__init__(self, value)
        self.signed = signed

    def _set_attributes(self):
        self.width = self.value.bit_length() + 1
        self.point = 0

    def _implement(self, m, seq, svalid=None, senable=None):
        data = vtypes.Int(self.value, width=self.width, signed=self.signed)
        self.sig_data = data


class Float(_Constant):

    def _set_attributes(self):
        self.width = 32
        self.point = 0
        self.signed = True


class FixedPoint(_Constant):

    def __init__(self, value, point=0, signed=True):
        _Constant.__init__(self, value)
        self.point = point
        self.signed = signed

    def _set_attributes(self):
        self.width = self.value.bit_length() + 1
        self.point = 0

    def _implement(self, m, seq, svalid=None, senable=None):
        data = vtypes.Int(self.value, width=self.width)
        self.sig_data = data


class Str(_Constant):

    def _set_attributes(self):
        self.width = 0
        self.point = 0
        self.signed = False


class Substream(_SpecialOperator):
    def __init__(self, substrm, strm=None):
        _SpecialOperator.__init__(self)
        self.strm = strm
        self._set_managers()

        self.width = 1
        self.point = 0
        self.signed = True

        self.graph_label = substrm.name if hasattr(substrm, 'name') else 'Substream'
        self.graph_shape = 'box'
        self.graph_peripheries = 2

        if not substrm.aswire:
            raise ValueError('aswire must be True.')
        if substrm.module is None:
            raise ValueError('module must not be None.')
        if substrm.clock is None:
            raise ValueError('clock must not be None.')
        if substrm.reset is None:
            raise ValueError('reset must not be None.')
        if (substrm.ivalid is not None or
                substrm.iready is not None):
            raise ValueError('ivalid and iready signals must be empty.')
        if (substrm.ovalid is not None or
                substrm.oready is not None):
            raise ValueError('ovalid and oready signals must be empty.')

        if not substrm.implemented:
            substrm.implement()

        self.substrm = substrm
        self.latency = substrm.pipeline_depth() + 1
        self.conds = OrderedDict()

    def _set_managers(self):
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def write(self, name, data, cond=None):
        wdata = _to_constant(data)
        self.args.append(wdata)
        if name in self.conds:
            raise ValueError('already assigned')
        self.conds[name] = cond

    def read(self, name):
        var = self.substrm.get_named_numeric(name)
        if self.strm is None:
            return _SubstreamOutput(self, var)
        return self.strm._SubstreamOutput(self, var, name)

    def _implement(self, m, seq, svalid=None, senable=None):
        arg_data = [arg.sig_data for arg in self.args]

        for data, (name, cond) in zip(arg_data, self.conds.items()):
            var = self.substrm.get_named_numeric(name)
            var.write(data, cond)

        self.sig_data = vtypes.Int(0)


class _SubstreamOutput(_UnaryOperator):

    def __init__(self, substrm, output_var, var_name):
        _UnaryOperator.__init__(self, substrm)
        self.var_name = var_name
        self.output_var = output_var

        self.width = output_var.bit_length()
        self.point = output_var.get_point()
        self.signed = output_var.get_signed()

        self.graph_label = substrm.graph_label + '\n' + self.var_name
        self.graph_shape = 'box'
        self.graph_peripheries = 2

    def _implement(self, m, seq, svalid=None, senable=None):
        width = self.bit_length()
        signed = self.get_signed()
        rdata = self.output_var.read()

        if self.latency == 0:
            data = m.Wire(self.name('data'), width, signed=signed)
            data.assign(rdata)
            self.sig_data = data

        elif self.latency == 1:
            data = m.Reg(self.name('data'), width, initval=0, signed=signed)
            self.sig_data = data
            seq(data(rdata), cond=senable)

        else:
            prev_data = None

            for i in range(self.latency):
                data = m.Reg(self.name('data_d%d' % i),
                             width, initval=0, signed=signed)
                if i == 0:
                    seq(data(rdata), cond=senable)
                else:
                    seq(data(prev_data), cond=senable)
                prev_data = data

            self.sig_data = data


class RingBuffer(_UnaryOperator):
    latency = 1

    def __init__(self, var, length,
                 enable=None, reset=None):

        self.enable = _to_constant(enable)
        if self.enable is not None:
            self.enable._add_sink(self)

        self.reset = _to_constant(reset)
        if self.reset is not None:
            self.reset._add_sink(self)

        self.length = length

        _UnaryOperator.__init__(self, var)

        self.num_ports = 1
        self.read_vars = []

        self.graph_label = 'RingBufferIn'
        self.graph_shape = 'box'

    def _set_managers(self):
        self._set_strm(_get_strm(self.right, self.enable, self.reset))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def read(self, offset):
        var = _RingBufferOutput(self, offset, self.num_ports,
                                self.enable, self.reset)
        self.read_vars.append(var)
        self.num_ports += 1
        return var

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        datawidth = self.bit_length()
        addrwidth = int(ceil(log(self.length, 2)))
        signed = self.get_signed()

        clk = m._clock
        self.ram = ram.SyncRAM(m, self.name('ram'),
                               clk, datawidth, addrwidth, self.num_ports)

        enabledata = self.enable.sig_data if self.enable is not None else None
        resetdata = self.reset.sig_data if self.reset is not None else None

        wdata = m.Wire(self.name('wdata'), datawidth, signed=signed)
        wdata.assign(self.right.sig_data)

        waddr = m.Reg(self.name('waddr'), addrwidth, initval=0)

        wcond = _and_vars(svalid, senable, enabledata)

        next_waddr = vtypes.Mux(waddr == self.length - 1, 0, waddr + 1)
        seq(waddr(next_waddr), cond=wcond)

        reset_waddr = 0
        reset_cond = _and_vars(svalid, senable, enabledata, resetdata)
        seq(waddr(reset_waddr), cond=reset_cond)

        resetdata_x = vtypes.Not(resetdata) if resetdata is not None else 1
        wenable = _and_vars(svalid, senable, enabledata, resetdata_x)
        self.ram.connect(0, waddr, wdata, wenable)

        self.sig_data = wdata


class _RingBufferOutput(_BinaryOperator):
    latency = 1

    def __init__(self, buf, offset, port,
                 enable=None, reset=None):

        self.enable = _to_constant(enable)
        if self.enable is not None:
            self.enable._add_sink(self)

        self.reset = _to_constant(reset)
        if self.reset is not None:
            self.reset._add_sink(self)

        self.port = port

        _BinaryOperator.__init__(self, buf, offset)
        self.buf = buf

        self.graph_label = 'RingBufferOut'
        self.graph_shape = 'box'

    def _set_managers(self):
        self._set_strm(_get_strm(self.left, self.right, self.enable, self.reset))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        datawidth = self.bit_length()
        addrwidth = int(ceil(log(self.buf.length, 2)))
        signed = self.get_signed()

        enabledata = self.enable.sig_data if self.enable is not None else None
        resetdata = self.reset.sig_data if self.reset is not None else None

        rdata = m.Wire(self.name('rdata'), datawidth, signed=signed)

        raddr_base = m.Reg(self.name('raddr'), addrwidth, initval=0)

        rcond = _and_vars(svalid, senable, enabledata)

        next_raddr_base = vtypes.Mux(raddr_base == self.buf.length - 1,
                                     0, raddr_base + 1)
        seq(raddr_base(next_raddr_base), cond=rcond)

        reset_raddr_base = 0
        reset_cond = _and_vars(svalid, senable, enabledata, resetdata)
        seq(raddr_base(reset_raddr_base), cond=reset_cond)

        raddr = raddr_base + self.right.sig_data
        raddr = vtypes.Mux(raddr >= self.buf.length,
                           raddr - self.buf.length, raddr)

        self.buf.ram.connect(self.port, raddr, 0, 0)
        rdata.assign(self.buf.ram.rdata(self.port))

        self.sig_data = rdata


class Scratchpad(_BinaryOperator):
    latency = 1

    def __init__(self, var, addr, length,
                 when=None, reset=None):

        self.enable = _to_constant(when)
        if self.enable is not None:
            self.enable._add_sink(self)

        self.reset = _to_constant(reset)
        if self.reset is not None:
            self.reset._add_sink(self)

        self.length = length

        _BinaryOperator.__init__(self, var, addr)

        self.num_ports = 1
        self.read_vars = []

        self.graph_label = 'ScratchpadIn'
        self.graph_shape = 'box'

    def _set_managers(self):
        self._set_strm(_get_strm(self.left, self.right, self.enable, self.reset))
        self._set_module(getattr(self.strm, 'module', None))
        self._set_seq(getattr(self.strm, 'seq', None))

    def read(self, addr):
        var = _ScratchpadOutput(self, addr, self.num_ports)
        self.read_vars.append(var)
        self.num_ports += 1
        return var

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        datawidth = self.bit_length()
        addrwidth = int(ceil(log(self.length, 2)))
        signed = self.get_signed()

        clk = m._clock
        self.ram = ram.SyncRAM(m, self.name('ram'),
                               clk, datawidth, addrwidth, self.num_ports)

        enabledata = self.enable.sig_data if self.enable is not None else None
        resetdata = self.reset.sig_data if self.reset is not None else None

        wdata = m.Wire(self.name('wdata'), datawidth, signed=signed)
        wdata.assign(self.left.sig_data)

        waddr = m.Wire(self.name('waddr'), addrwidth)
        waddr.assign(self.right.sig_data)

        resetdata_x = vtypes.Not(resetdata) if resetdata is not None else 1
        wenable = _and_vars(svalid, senable, enabledata, resetdata_x)
        self.ram.connect(0, waddr, wdata, wenable)

        self.sig_data = wdata


class _ScratchpadOutput(_BinaryOperator):
    latency = 1

    def __init__(self, sp, addr, port):
        self.port = port
        _BinaryOperator.__init__(self, sp, addr)
        self.sp = sp

        self.graph_label = 'ScratchpadOut'
        self.graph_shape = 'box'

    def _implement(self, m, seq, svalid=None, senable=None):
        if self.latency != 1:
            raise ValueError("Latency mismatch '%d' vs '%s'" %
                             (self.latency, 1))

        datawidth = self.bit_length()
        addrwidth = int(ceil(log(self.sp.length, 2)))
        signed = self.get_signed()

        rdata = m.Wire(self.name('rdata'), datawidth, signed=signed)

        raddr = m.Wire(self.name('raddr'), addrwidth)
        raddr.assign(self.right.sig_data)

        self.sp.ram.connect(self.port, raddr, 0, 0)
        rdata.assign(self.sp.ram.rdata(self.port))

        self.sig_data = rdata


def make_condition(*cond, **kwargs):
    ready = kwargs['ready'] if 'ready' in kwargs else None

    _cond = []
    for c in cond:
        if isinstance(c, (tuple, list)):
            _cond.extend(c)
        else:
            _cond.append(c)

    cond = _cond

    new_cond = []
    for c in cond:
        if c is None:
            continue
        if isinstance(c, _Numeric):
            d, v = c.read(cond=ready)
            new_cond.append(d)
            new_cond.append(v)
        else:
            new_cond.append(c)

    return _make_condition(*new_cond)


def is_stream_object(*objs):
    for obj in objs:
        if isinstance(obj, _Node):
            return True
    return False


def _to_constant(obj):
    if isinstance(obj, (int, float, bool, str)):
        return Constant(obj)
    if isinstance(obj, vtypes._Numeric):
        return _from_vtypes_value(obj)
    return obj


def _get_strm(*vars):
    ret = None
    for var in vars:
        v = getattr(var, 'strm', None)
        if v is None:
            continue
        if ret is None:
            ret = v
            continue
        if v.object_id < ret.object_id:
            if v.module != ret.module:
                raise ValueError("Different modules")
            if id(v.clock) != id(ret.clock):
                raise ValueError("Different clock domains: '%s' and '%s'" %
                                 (str(v.clock), str(ret.clock)))
            if id(v.reset) != id(ret.reset):
                raise ValueError("Different reset domains: '%s' and '%s'" %
                                 (str(v.reset), str(ret.reset)))
            ret = v
    return ret


def _max(*vars):
    m = None
    for v in vars:
        if v is None:
            continue
        if m is None:
            m = v
            continue
        if m < v:
            m = v
    return m


def _and_vars(*vars):
    ret = None
    for var in vars:
        if var is None:
            continue
        if ret is None:
            ret = var
        else:
            ret = vtypes.AndList(ret, var)
    return ret


def _from_vtypes_value(value):
    if isinstance(value, vtypes.Int):
        if not isinstance(value.value, int):
            raise TypeError("Unsupported type for Constant '%s'" %
                            str(type(value)))
        return Int(value.value)

    if isinstance(value, vtypes.Float):
        return Float(value.value)

    if isinstance(value, vtypes.Str):
        return Str(value.value)

    if isinstance(value, vtypes._Numeric):
        return ParameterVariable(value)

    raise TypeError("Unsupported type '%s'" % str(type(value)))



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import seq_countup_noreset

expected_verilog = """
module test #
(
  parameter INTERVAL = 16
);

  reg CLK;
  reg RST;
  wire [(8 - 1):0] LED;

  blinkled
  #(
    .INTERVAL(INTERVAL)
  )
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = (!CLK);
    end
  end

  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    $finish;
  end
endmodule

module blinkled #
(
  parameter INTERVAL = 16
)
(
  input CLK,
  input RST,
  output reg [(8 - 1):0] LED
);

  reg [(32 - 1):0] count;

  always @(posedge CLK) begin
    $display("LED:%d count:%d", LED, count);
    if((count < (INTERVAL - 1))) begin
      count <= (count + 1);
    end 
    if((count == (INTERVAL - 1))) begin
      count <= 0;
    end 
    if((count == (INTERVAL - 1))) begin
      LED <= (LED + 1);
    end 
  end
endmodule
"""
def test():
    veriloggen.reset()
    test_module = seq_countup_noreset.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkSub():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    count = m.OutputReg('count', 32)
    m.Always(Posedge(clk))(
        If(rst)(
            count(0)
        ).Else(
            If(count == 1023)(
                count(0)
            ).Else(
                count(count + 1)
            )
        ))
    return m

def mkLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    count = m.Wire('count', 32)

    sub = mkSub()
    m.Instance(sub, 'inst_sub', m.connect_params(sub), m.connect_ports(sub))
    
    # by multiple definition, throws an exception here
    m.Instance(sub, 'inst_sub', m.connect_params(sub), m.connect_ports(sub))
    
    m.Always(Posedge(clk))(
        If(rst)(
            led(0)
        ).Else(
            If(count == 1023)(
                led(led + 1)
            )
        ))

    return m

if __name__ == '__main__':
    try:
        led = mkLed()
    except ValueError as e:
        print(e.args[0])
        print('But it was detected')
        sys.exit()

    raise ValueError("Multiple definition was not detected.")

    #verilog = led.to_verilog()
    #print(verilog)



"""
   Veriloggen: A library for constructing a Verilog HDL source code in Python

   Copyright 2015, Shinya Takamaeda-Yamazaki and Contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from __future__ import absolute_import
from __future__ import print_function

# expand the recursive call limit for python 3.6 and later
import sys
sys.setrecursionlimit(1000 * 10)

# Verilog HDL Core
from .core.vtypes import *
from .core.module import Module, StubModule, Instance, GenerateFor, GenerateIf, connect_same_name
from .core.function import Function, FunctionCall
from .core.task import Task, TaskCall
from .core.submodule import Submodule

# Code Generator
from .verilog import from_verilog

# Verilog Simulator
from .simulation import simulation

# Abstract Extension
from .seq.seq import Seq, TmpSeq, make_condition
from .fsm.fsm import FSM, TmpFSM
from .pipeline.pipeline import Pipeline

# Extension reset
from .seq import reset as seq_reset
from .fsm import reset as fsm_reset
from .dataflow import reset as dataflow_reset
from .stream import reset as stream_reset
from .thread import reset as thread_reset
from .types import reset as types_reset


def reset():
    seq_reset()
    fsm_reset()
    dataflow_reset()
    stream_reset()
    thread_reset()
    types_reset()



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import simulation_simulator_verilator
from veriloggen import *

expected_rslt = """\
LED:  0 count:         0
LED:  0 count:         1
LED:  0 count:         2
LED:  0 count:         3
LED:  0 count:         4
LED:  0 count:         5
LED:  0 count:         6
LED:  0 count:         7
LED:  0 count:         8
LED:  0 count:         9
LED:  0 count:        10
LED:  0 count:         0
LED:  0 count:         1
LED:  0 count:         2
LED:  0 count:         3
LED:  0 count:         4
LED:  0 count:         5
LED:  0 count:         6
LED:  0 count:         7
LED:  0 count:         8
LED:  0 count:         9
LED:  0 count:        10
LED:  0 count:        11
LED:  0 count:        12
LED:  0 count:        13
LED:  0 count:        14
LED:  0 count:        15
LED:  0 count:        16
LED:  0 count:        17
LED:  0 count:        18
LED:  0 count:        19
LED:  0 count:        20
LED:  0 count:        21
LED:  0 count:        22
LED:  0 count:        23
LED:  0 count:        24
LED:  0 count:        25
LED:  0 count:        26
LED:  0 count:        27
LED:  0 count:        28
LED:  0 count:        29
LED:  0 count:        30
LED:  0 count:        31
LED:  0 count:        32
LED:  0 count:        33
LED:  0 count:        34
LED:  0 count:        35
LED:  0 count:        36
LED:  0 count:        37
LED:  0 count:        38
LED:  0 count:        39
LED:  0 count:        40
LED:  0 count:        41
LED:  0 count:        42
LED:  0 count:        43
LED:  0 count:        44
LED:  0 count:        45
LED:  0 count:        46
LED:  0 count:        47
LED:  0 count:        48
LED:  0 count:        49
LED:  0 count:        50
LED:  0 count:        51
LED:  0 count:        52
LED:  0 count:        53
LED:  0 count:        54
LED:  0 count:        55
LED:  0 count:        56
LED:  0 count:        57
LED:  0 count:        58
LED:  0 count:        59
LED:  0 count:        60
LED:  0 count:        61
LED:  0 count:        62
LED:  0 count:        63
LED:  0 count:        64
LED:  0 count:        65
LED:  0 count:        66
LED:  0 count:        67
LED:  0 count:        68
LED:  0 count:        69
LED:  0 count:        70
LED:  0 count:        71
LED:  0 count:        72
LED:  0 count:        73
LED:  0 count:        74
LED:  0 count:        75
LED:  0 count:        76
LED:  0 count:        77
LED:  0 count:        78
LED:  0 count:        79
LED:  0 count:        80
LED:  0 count:        81
LED:  0 count:        82
LED:  0 count:        83
LED:  0 count:        84
LED:  0 count:        85
LED:  0 count:        86
LED:  0 count:        87
LED:  0 count:        88
LED:  0 count:        89
LED:  0 count:        90
LED:  0 count:        91
LED:  0 count:        92
LED:  0 count:        93
LED:  0 count:        94
LED:  0 count:        95
LED:  0 count:        96
LED:  0 count:        97
LED:  0 count:        98
LED:  0 count:        99
LED:  0 count:       100
LED:  0 count:       101
LED:  0 count:       102
LED:  0 count:       103
LED:  0 count:       104
LED:  0 count:       105
LED:  0 count:       106
LED:  0 count:       107
LED:  0 count:       108
LED:  0 count:       109
LED:  0 count:       110
LED:  0 count:       111
LED:  0 count:       112
LED:  0 count:       113
LED:  0 count:       114
LED:  0 count:       115
LED:  0 count:       116
LED:  0 count:       117
LED:  0 count:       118
LED:  0 count:       119
LED:  0 count:       120
LED:  0 count:       121
LED:  0 count:       122
LED:  0 count:       123
LED:  0 count:       124
LED:  0 count:       125
LED:  0 count:       126
LED:  0 count:       127
LED:  0 count:       128
LED:  0 count:       129
LED:  0 count:       130
LED:  0 count:       131
LED:  0 count:       132
LED:  0 count:       133
LED:  0 count:       134
LED:  0 count:       135
LED:  0 count:       136
LED:  0 count:       137
LED:  0 count:       138
LED:  0 count:       139
LED:  0 count:       140
LED:  0 count:       141
LED:  0 count:       142
LED:  0 count:       143
LED:  0 count:       144
LED:  0 count:       145
LED:  0 count:       146
LED:  0 count:       147
LED:  0 count:       148
LED:  0 count:       149
LED:  0 count:       150
LED:  0 count:       151
LED:  0 count:       152
LED:  0 count:       153
LED:  0 count:       154
LED:  0 count:       155
LED:  0 count:       156
LED:  0 count:       157
LED:  0 count:       158
LED:  0 count:       159
LED:  0 count:       160
LED:  0 count:       161
LED:  0 count:       162
LED:  0 count:       163
LED:  0 count:       164
LED:  0 count:       165
LED:  0 count:       166
LED:  0 count:       167
LED:  0 count:       168
LED:  0 count:       169
LED:  0 count:       170
LED:  0 count:       171
LED:  0 count:       172
LED:  0 count:       173
LED:  0 count:       174
LED:  0 count:       175
LED:  0 count:       176
LED:  0 count:       177
LED:  0 count:       178
LED:  0 count:       179
LED:  0 count:       180
LED:  0 count:       181
LED:  0 count:       182
LED:  0 count:       183
LED:  0 count:       184
LED:  0 count:       185
LED:  0 count:       186
LED:  0 count:       187
LED:  0 count:       188
LED:  0 count:       189
LED:  0 count:       190
LED:  0 count:       191
LED:  0 count:       192
LED:  0 count:       193
LED:  0 count:       194
LED:  0 count:       195
LED:  0 count:       196
LED:  0 count:       197
LED:  0 count:       198
LED:  0 count:       199
LED:  0 count:       200
LED:  0 count:       201
LED:  0 count:       202
LED:  0 count:       203
LED:  0 count:       204
LED:  0 count:       205
LED:  0 count:       206
LED:  0 count:       207
LED:  0 count:       208
LED:  0 count:       209
LED:  0 count:       210
LED:  0 count:       211
LED:  0 count:       212
LED:  0 count:       213
LED:  0 count:       214
LED:  0 count:       215
LED:  0 count:       216
LED:  0 count:       217
LED:  0 count:       218
LED:  0 count:       219
LED:  0 count:       220
LED:  0 count:       221
LED:  0 count:       222
LED:  0 count:       223
LED:  0 count:       224
LED:  0 count:       225
LED:  0 count:       226
LED:  0 count:       227
LED:  0 count:       228
LED:  0 count:       229
LED:  0 count:       230
LED:  0 count:       231
LED:  0 count:       232
LED:  0 count:       233
LED:  0 count:       234
LED:  0 count:       235
LED:  0 count:       236
LED:  0 count:       237
LED:  0 count:       238
LED:  0 count:       239
LED:  0 count:       240
LED:  0 count:       241
LED:  0 count:       242
LED:  0 count:       243
LED:  0 count:       244
LED:  0 count:       245
LED:  0 count:       246
LED:  0 count:       247
LED:  0 count:       248
LED:  0 count:       249
LED:  0 count:       250
LED:  0 count:       251
LED:  0 count:       252
LED:  0 count:       253
LED:  0 count:       254
LED:  0 count:       255
LED:  0 count:       256
LED:  0 count:       257
LED:  0 count:       258
LED:  0 count:       259
LED:  0 count:       260
LED:  0 count:       261
LED:  0 count:       262
LED:  0 count:       263
LED:  0 count:       264
LED:  0 count:       265
LED:  0 count:       266
LED:  0 count:       267
LED:  0 count:       268
LED:  0 count:       269
LED:  0 count:       270
LED:  0 count:       271
LED:  0 count:       272
LED:  0 count:       273
LED:  0 count:       274
LED:  0 count:       275
LED:  0 count:       276
LED:  0 count:       277
LED:  0 count:       278
LED:  0 count:       279
LED:  0 count:       280
LED:  0 count:       281
LED:  0 count:       282
LED:  0 count:       283
LED:  0 count:       284
LED:  0 count:       285
LED:  0 count:       286
LED:  0 count:       287
LED:  0 count:       288
LED:  0 count:       289
LED:  0 count:       290
LED:  0 count:       291
LED:  0 count:       292
LED:  0 count:       293
LED:  0 count:       294
LED:  0 count:       295
LED:  0 count:       296
LED:  0 count:       297
LED:  0 count:       298
LED:  0 count:       299
LED:  0 count:       300
LED:  0 count:       301
LED:  0 count:       302
LED:  0 count:       303
LED:  0 count:       304
LED:  0 count:       305
LED:  0 count:       306
LED:  0 count:       307
LED:  0 count:       308
LED:  0 count:       309
LED:  0 count:       310
LED:  0 count:       311
LED:  0 count:       312
LED:  0 count:       313
LED:  0 count:       314
LED:  0 count:       315
LED:  0 count:       316
LED:  0 count:       317
LED:  0 count:       318
LED:  0 count:       319
LED:  0 count:       320
LED:  0 count:       321
LED:  0 count:       322
LED:  0 count:       323
LED:  0 count:       324
LED:  0 count:       325
LED:  0 count:       326
LED:  0 count:       327
LED:  0 count:       328
LED:  0 count:       329
LED:  0 count:       330
LED:  0 count:       331
LED:  0 count:       332
LED:  0 count:       333
LED:  0 count:       334
LED:  0 count:       335
LED:  0 count:       336
LED:  0 count:       337
LED:  0 count:       338
LED:  0 count:       339
LED:  0 count:       340
LED:  0 count:       341
LED:  0 count:       342
LED:  0 count:       343
LED:  0 count:       344
LED:  0 count:       345
LED:  0 count:       346
LED:  0 count:       347
LED:  0 count:       348
LED:  0 count:       349
LED:  0 count:       350
LED:  0 count:       351
LED:  0 count:       352
LED:  0 count:       353
LED:  0 count:       354
LED:  0 count:       355
LED:  0 count:       356
LED:  0 count:       357
LED:  0 count:       358
LED:  0 count:       359
LED:  0 count:       360
LED:  0 count:       361
LED:  0 count:       362
LED:  0 count:       363
LED:  0 count:       364
LED:  0 count:       365
LED:  0 count:       366
LED:  0 count:       367
LED:  0 count:       368
LED:  0 count:       369
LED:  0 count:       370
LED:  0 count:       371
LED:  0 count:       372
LED:  0 count:       373
LED:  0 count:       374
LED:  0 count:       375
LED:  0 count:       376
LED:  0 count:       377
LED:  0 count:       378
LED:  0 count:       379
LED:  0 count:       380
LED:  0 count:       381
LED:  0 count:       382
LED:  0 count:       383
LED:  0 count:       384
LED:  0 count:       385
LED:  0 count:       386
LED:  0 count:       387
LED:  0 count:       388
LED:  0 count:       389
LED:  0 count:       390
LED:  0 count:       391
LED:  0 count:       392
LED:  0 count:       393
LED:  0 count:       394
LED:  0 count:       395
LED:  0 count:       396
LED:  0 count:       397
LED:  0 count:       398
LED:  0 count:       399
LED:  0 count:       400
LED:  0 count:       401
LED:  0 count:       402
LED:  0 count:       403
LED:  0 count:       404
LED:  0 count:       405
LED:  0 count:       406
LED:  0 count:       407
LED:  0 count:       408
LED:  0 count:       409
LED:  0 count:       410
LED:  0 count:       411
LED:  0 count:       412
LED:  0 count:       413
LED:  0 count:       414
LED:  0 count:       415
LED:  0 count:       416
LED:  0 count:       417
LED:  0 count:       418
LED:  0 count:       419
LED:  0 count:       420
LED:  0 count:       421
LED:  0 count:       422
LED:  0 count:       423
LED:  0 count:       424
LED:  0 count:       425
LED:  0 count:       426
LED:  0 count:       427
LED:  0 count:       428
LED:  0 count:       429
LED:  0 count:       430
LED:  0 count:       431
LED:  0 count:       432
LED:  0 count:       433
LED:  0 count:       434
LED:  0 count:       435
LED:  0 count:       436
LED:  0 count:       437
LED:  0 count:       438
LED:  0 count:       439
LED:  0 count:       440
LED:  0 count:       441
LED:  0 count:       442
LED:  0 count:       443
LED:  0 count:       444
LED:  0 count:       445
LED:  0 count:       446
LED:  0 count:       447
LED:  0 count:       448
LED:  0 count:       449
LED:  0 count:       450
LED:  0 count:       451
LED:  0 count:       452
LED:  0 count:       453
LED:  0 count:       454
LED:  0 count:       455
LED:  0 count:       456
LED:  0 count:       457
LED:  0 count:       458
LED:  0 count:       459
LED:  0 count:       460
LED:  0 count:       461
LED:  0 count:       462
LED:  0 count:       463
LED:  0 count:       464
LED:  0 count:       465
LED:  0 count:       466
LED:  0 count:       467
LED:  0 count:       468
LED:  0 count:       469
LED:  0 count:       470
LED:  0 count:       471
LED:  0 count:       472
LED:  0 count:       473
LED:  0 count:       474
LED:  0 count:       475
LED:  0 count:       476
LED:  0 count:       477
LED:  0 count:       478
LED:  0 count:       479
LED:  0 count:       480
LED:  0 count:       481
LED:  0 count:       482
LED:  0 count:       483
LED:  0 count:       484
LED:  0 count:       485
LED:  0 count:       486
LED:  0 count:       487
LED:  0 count:       488
LED:  0 count:       489
LED:  0 count:       490
LED:  0 count:       491
LED:  0 count:       492
LED:  0 count:       493
LED:  0 count:       494
LED:  0 count:       495
LED:  0 count:       496
LED:  0 count:       497
LED:  0 count:       498
LED:  0 count:       499
LED:  0 count:       500
LED:  0 count:       501
LED:  0 count:       502
LED:  0 count:       503
LED:  0 count:       504
LED:  0 count:       505
LED:  0 count:       506
LED:  0 count:       507
LED:  0 count:       508
LED:  0 count:       509
LED:  0 count:       510
LED:  0 count:       511
LED:  0 count:       512
LED:  0 count:       513
LED:  0 count:       514
LED:  0 count:       515
LED:  0 count:       516
LED:  0 count:       517
LED:  0 count:       518
LED:  0 count:       519
LED:  0 count:       520
LED:  0 count:       521
LED:  0 count:       522
LED:  0 count:       523
LED:  0 count:       524
LED:  0 count:       525
LED:  0 count:       526
LED:  0 count:       527
LED:  0 count:       528
LED:  0 count:       529
LED:  0 count:       530
LED:  0 count:       531
LED:  0 count:       532
LED:  0 count:       533
LED:  0 count:       534
LED:  0 count:       535
LED:  0 count:       536
LED:  0 count:       537
LED:  0 count:       538
LED:  0 count:       539
LED:  0 count:       540
LED:  0 count:       541
LED:  0 count:       542
LED:  0 count:       543
LED:  0 count:       544
LED:  0 count:       545
LED:  0 count:       546
LED:  0 count:       547
LED:  0 count:       548
LED:  0 count:       549
LED:  0 count:       550
LED:  0 count:       551
LED:  0 count:       552
LED:  0 count:       553
LED:  0 count:       554
LED:  0 count:       555
LED:  0 count:       556
LED:  0 count:       557
LED:  0 count:       558
LED:  0 count:       559
LED:  0 count:       560
LED:  0 count:       561
LED:  0 count:       562
LED:  0 count:       563
LED:  0 count:       564
LED:  0 count:       565
LED:  0 count:       566
LED:  0 count:       567
LED:  0 count:       568
LED:  0 count:       569
LED:  0 count:       570
LED:  0 count:       571
LED:  0 count:       572
LED:  0 count:       573
LED:  0 count:       574
LED:  0 count:       575
LED:  0 count:       576
LED:  0 count:       577
LED:  0 count:       578
LED:  0 count:       579
LED:  0 count:       580
LED:  0 count:       581
LED:  0 count:       582
LED:  0 count:       583
LED:  0 count:       584
LED:  0 count:       585
LED:  0 count:       586
LED:  0 count:       587
LED:  0 count:       588
LED:  0 count:       589
LED:  0 count:       590
LED:  0 count:       591
LED:  0 count:       592
LED:  0 count:       593
LED:  0 count:       594
LED:  0 count:       595
LED:  0 count:       596
LED:  0 count:       597
LED:  0 count:       598
LED:  0 count:       599
LED:  0 count:       600
LED:  0 count:       601
LED:  0 count:       602
LED:  0 count:       603
LED:  0 count:       604
LED:  0 count:       605
LED:  0 count:       606
LED:  0 count:       607
LED:  0 count:       608
LED:  0 count:       609
LED:  0 count:       610
LED:  0 count:       611
LED:  0 count:       612
LED:  0 count:       613
LED:  0 count:       614
LED:  0 count:       615
LED:  0 count:       616
LED:  0 count:       617
LED:  0 count:       618
LED:  0 count:       619
LED:  0 count:       620
LED:  0 count:       621
LED:  0 count:       622
LED:  0 count:       623
LED:  0 count:       624
LED:  0 count:       625
LED:  0 count:       626
LED:  0 count:       627
LED:  0 count:       628
LED:  0 count:       629
LED:  0 count:       630
LED:  0 count:       631
LED:  0 count:       632
LED:  0 count:       633
LED:  0 count:       634
LED:  0 count:       635
LED:  0 count:       636
LED:  0 count:       637
LED:  0 count:       638
LED:  0 count:       639
LED:  0 count:       640
LED:  0 count:       641
LED:  0 count:       642
LED:  0 count:       643
LED:  0 count:       644
LED:  0 count:       645
LED:  0 count:       646
LED:  0 count:       647
LED:  0 count:       648
LED:  0 count:       649
LED:  0 count:       650
LED:  0 count:       651
LED:  0 count:       652
LED:  0 count:       653
LED:  0 count:       654
LED:  0 count:       655
LED:  0 count:       656
LED:  0 count:       657
LED:  0 count:       658
LED:  0 count:       659
LED:  0 count:       660
LED:  0 count:       661
LED:  0 count:       662
LED:  0 count:       663
LED:  0 count:       664
LED:  0 count:       665
LED:  0 count:       666
LED:  0 count:       667
LED:  0 count:       668
LED:  0 count:       669
LED:  0 count:       670
LED:  0 count:       671
LED:  0 count:       672
LED:  0 count:       673
LED:  0 count:       674
LED:  0 count:       675
LED:  0 count:       676
LED:  0 count:       677
LED:  0 count:       678
LED:  0 count:       679
LED:  0 count:       680
LED:  0 count:       681
LED:  0 count:       682
LED:  0 count:       683
LED:  0 count:       684
LED:  0 count:       685
LED:  0 count:       686
LED:  0 count:       687
LED:  0 count:       688
LED:  0 count:       689
LED:  0 count:       690
LED:  0 count:       691
LED:  0 count:       692
LED:  0 count:       693
LED:  0 count:       694
LED:  0 count:       695
LED:  0 count:       696
LED:  0 count:       697
LED:  0 count:       698
LED:  0 count:       699
LED:  0 count:       700
LED:  0 count:       701
LED:  0 count:       702
LED:  0 count:       703
LED:  0 count:       704
LED:  0 count:       705
LED:  0 count:       706
LED:  0 count:       707
LED:  0 count:       708
LED:  0 count:       709
LED:  0 count:       710
LED:  0 count:       711
LED:  0 count:       712
LED:  0 count:       713
LED:  0 count:       714
LED:  0 count:       715
LED:  0 count:       716
LED:  0 count:       717
LED:  0 count:       718
LED:  0 count:       719
LED:  0 count:       720
LED:  0 count:       721
LED:  0 count:       722
LED:  0 count:       723
LED:  0 count:       724
LED:  0 count:       725
LED:  0 count:       726
LED:  0 count:       727
LED:  0 count:       728
LED:  0 count:       729
LED:  0 count:       730
LED:  0 count:       731
LED:  0 count:       732
LED:  0 count:       733
LED:  0 count:       734
LED:  0 count:       735
LED:  0 count:       736
LED:  0 count:       737
LED:  0 count:       738
LED:  0 count:       739
LED:  0 count:       740
LED:  0 count:       741
LED:  0 count:       742
LED:  0 count:       743
LED:  0 count:       744
LED:  0 count:       745
LED:  0 count:       746
LED:  0 count:       747
LED:  0 count:       748
LED:  0 count:       749
LED:  0 count:       750
LED:  0 count:       751
LED:  0 count:       752
LED:  0 count:       753
LED:  0 count:       754
LED:  0 count:       755
LED:  0 count:       756
LED:  0 count:       757
LED:  0 count:       758
LED:  0 count:       759
LED:  0 count:       760
LED:  0 count:       761
LED:  0 count:       762
LED:  0 count:       763
LED:  0 count:       764
LED:  0 count:       765
LED:  0 count:       766
LED:  0 count:       767
LED:  0 count:       768
LED:  0 count:       769
LED:  0 count:       770
LED:  0 count:       771
LED:  0 count:       772
LED:  0 count:       773
LED:  0 count:       774
LED:  0 count:       775
LED:  0 count:       776
LED:  0 count:       777
LED:  0 count:       778
LED:  0 count:       779
LED:  0 count:       780
LED:  0 count:       781
LED:  0 count:       782
LED:  0 count:       783
LED:  0 count:       784
LED:  0 count:       785
LED:  0 count:       786
LED:  0 count:       787
LED:  0 count:       788
LED:  0 count:       789
LED:  0 count:       790
LED:  0 count:       791
LED:  0 count:       792
LED:  0 count:       793
LED:  0 count:       794
LED:  0 count:       795
LED:  0 count:       796
LED:  0 count:       797
LED:  0 count:       798
LED:  0 count:       799
LED:  0 count:       800
LED:  0 count:       801
LED:  0 count:       802
LED:  0 count:       803
LED:  0 count:       804
LED:  0 count:       805
LED:  0 count:       806
LED:  0 count:       807
LED:  0 count:       808
LED:  0 count:       809
LED:  0 count:       810
LED:  0 count:       811
LED:  0 count:       812
LED:  0 count:       813
LED:  0 count:       814
LED:  0 count:       815
LED:  0 count:       816
LED:  0 count:       817
LED:  0 count:       818
LED:  0 count:       819
LED:  0 count:       820
LED:  0 count:       821
LED:  0 count:       822
LED:  0 count:       823
LED:  0 count:       824
LED:  0 count:       825
LED:  0 count:       826
LED:  0 count:       827
LED:  0 count:       828
LED:  0 count:       829
LED:  0 count:       830
LED:  0 count:       831
LED:  0 count:       832
LED:  0 count:       833
LED:  0 count:       834
LED:  0 count:       835
LED:  0 count:       836
LED:  0 count:       837
LED:  0 count:       838
LED:  0 count:       839
LED:  0 count:       840
LED:  0 count:       841
LED:  0 count:       842
LED:  0 count:       843
LED:  0 count:       844
LED:  0 count:       845
LED:  0 count:       846
LED:  0 count:       847
LED:  0 count:       848
LED:  0 count:       849
LED:  0 count:       850
LED:  0 count:       851
LED:  0 count:       852
LED:  0 count:       853
LED:  0 count:       854
LED:  0 count:       855
LED:  0 count:       856
LED:  0 count:       857
LED:  0 count:       858
LED:  0 count:       859
LED:  0 count:       860
LED:  0 count:       861
LED:  0 count:       862
LED:  0 count:       863
LED:  0 count:       864
LED:  0 count:       865
LED:  0 count:       866
LED:  0 count:       867
LED:  0 count:       868
LED:  0 count:       869
LED:  0 count:       870
LED:  0 count:       871
LED:  0 count:       872
LED:  0 count:       873
LED:  0 count:       874
LED:  0 count:       875
LED:  0 count:       876
LED:  0 count:       877
LED:  0 count:       878
LED:  0 count:       879
LED:  0 count:       880
LED:  0 count:       881
LED:  0 count:       882
LED:  0 count:       883
LED:  0 count:       884
LED:  0 count:       885
LED:  0 count:       886
LED:  0 count:       887
LED:  0 count:       888
LED:  0 count:       889
LED:  0 count:       890
LED:  0 count:       891
LED:  0 count:       892
LED:  0 count:       893
LED:  0 count:       894
LED:  0 count:       895
LED:  0 count:       896
LED:  0 count:       897
LED:  0 count:       898
LED:  0 count:       899
LED:  0 count:       900
LED:  0 count:       901
LED:  0 count:       902
LED:  0 count:       903
LED:  0 count:       904
LED:  0 count:       905
LED:  0 count:       906
LED:  0 count:       907
LED:  0 count:       908
LED:  0 count:       909
LED:  0 count:       910
LED:  0 count:       911
LED:  0 count:       912
LED:  0 count:       913
LED:  0 count:       914
LED:  0 count:       915
LED:  0 count:       916
LED:  0 count:       917
LED:  0 count:       918
LED:  0 count:       919
LED:  0 count:       920
LED:  0 count:       921
LED:  0 count:       922
LED:  0 count:       923
LED:  0 count:       924
LED:  0 count:       925
LED:  0 count:       926
LED:  0 count:       927
LED:  0 count:       928
LED:  0 count:       929
LED:  0 count:       930
LED:  0 count:       931
LED:  0 count:       932
LED:  0 count:       933
LED:  0 count:       934
LED:  0 count:       935
LED:  0 count:       936
LED:  0 count:       937
LED:  0 count:       938
LED:  0 count:       939
LED:  0 count:       940
LED:  0 count:       941
LED:  0 count:       942
LED:  0 count:       943
LED:  0 count:       944
LED:  0 count:       945
LED:  0 count:       946
LED:  0 count:       947
LED:  0 count:       948
LED:  0 count:       949
LED:  0 count:       950
LED:  0 count:       951
LED:  0 count:       952
LED:  0 count:       953
LED:  0 count:       954
LED:  0 count:       955
LED:  0 count:       956
LED:  0 count:       957
LED:  0 count:       958
LED:  0 count:       959
LED:  0 count:       960
LED:  0 count:       961
LED:  0 count:       962
LED:  0 count:       963
LED:  0 count:       964
LED:  0 count:       965
LED:  0 count:       966
LED:  0 count:       967
LED:  0 count:       968
LED:  0 count:       969
LED:  0 count:       970
LED:  0 count:       971
LED:  0 count:       972
LED:  0 count:       973
LED:  0 count:       974
LED:  0 count:       975
LED:  0 count:       976
LED:  0 count:       977
LED:  0 count:       978
LED:  0 count:       979
LED:  0 count:       980
LED:  0 count:       981
LED:  0 count:       982
LED:  0 count:       983
LED:  0 count:       984
LED:  0 count:       985
LED:  0 count:       986
LED:  0 count:       987
LED:  0 count:       988
LED:  0 count:       989
LED:  0 count:       990
LED:  0 count:       991
LED:  0 count:       992
LED:  0 count:       993
LED:  0 count:       994
LED:  0 count:       995
LED:  0 count:       996
LED:  0 count:       997
LED:  0 count:       998
LED:  0 count:       999
LED:  0 count:      1000
LED:  0 count:      1001
LED:  0 count:      1002
LED:  0 count:      1003
LED:  0 count:      1004
LED:  0 count:      1005
LED:  0 count:      1006
LED:  0 count:      1007
LED:  0 count:      1008
LED:  0 count:      1009
LED:  0 count:      1010
LED:  0 count:      1011
LED:  0 count:      1012
LED:  0 count:      1013
LED:  0 count:      1014
LED:  0 count:      1015
LED:  0 count:      1016
LED:  0 count:      1017
LED:  0 count:      1018
LED:  0 count:      1019
LED:  0 count:      1020
LED:  0 count:      1021
LED:  0 count:      1022
LED:  0 count:      1023
LED:  1 count:         0
LED:  1 count:         1
LED:  1 count:         2
LED:  1 count:         3
LED:  1 count:         4
LED:  1 count:         5
LED:  1 count:         6
LED:  1 count:         7
LED:  1 count:         8
LED:  1 count:         9
LED:  1 count:        10
LED:  1 count:        11
LED:  1 count:        12
LED:  1 count:        13
LED:  1 count:        14
LED:  1 count:        15
LED:  1 count:        16
LED:  1 count:        17
LED:  1 count:        18
LED:  1 count:        19
LED:  1 count:        20
LED:  1 count:        21
LED:  1 count:        22
LED:  1 count:        23
LED:  1 count:        24
LED:  1 count:        25
LED:  1 count:        26
LED:  1 count:        27
LED:  1 count:        28
LED:  1 count:        29
LED:  1 count:        30
LED:  1 count:        31
LED:  1 count:        32
LED:  1 count:        33
LED:  1 count:        34
LED:  1 count:        35
LED:  1 count:        36
LED:  1 count:        37
LED:  1 count:        38
LED:  1 count:        39
LED:  1 count:        40
LED:  1 count:        41
LED:  1 count:        42
LED:  1 count:        43
LED:  1 count:        44
LED:  1 count:        45
LED:  1 count:        46
LED:  1 count:        47
LED:  1 count:        48
LED:  1 count:        49
LED:  1 count:        50
LED:  1 count:        51
LED:  1 count:        52
LED:  1 count:        53
LED:  1 count:        54
LED:  1 count:        55
LED:  1 count:        56
LED:  1 count:        57
LED:  1 count:        58
LED:  1 count:        59
LED:  1 count:        60
LED:  1 count:        61
LED:  1 count:        62
LED:  1 count:        63
LED:  1 count:        64
LED:  1 count:        65
LED:  1 count:        66
LED:  1 count:        67
LED:  1 count:        68
LED:  1 count:        69
LED:  1 count:        70
LED:  1 count:        71
LED:  1 count:        72
LED:  1 count:        73
LED:  1 count:        74
LED:  1 count:        75
LED:  1 count:        76
LED:  1 count:        77
LED:  1 count:        78
LED:  1 count:        79
LED:  1 count:        80
LED:  1 count:        81
LED:  1 count:        82
LED:  1 count:        83
LED:  1 count:        84
LED:  1 count:        85
LED:  1 count:        86
LED:  1 count:        87
LED:  1 count:        88
LED:  1 count:        89
LED:  1 count:        90
LED:  1 count:        91
LED:  1 count:        92
LED:  1 count:        93
LED:  1 count:        94
LED:  1 count:        95
LED:  1 count:        96
LED:  1 count:        97
LED:  1 count:        98
LED:  1 count:        99
LED:  1 count:       100
LED:  1 count:       101
LED:  1 count:       102
LED:  1 count:       103
LED:  1 count:       104
LED:  1 count:       105
LED:  1 count:       106
LED:  1 count:       107
LED:  1 count:       108
LED:  1 count:       109
LED:  1 count:       110
LED:  1 count:       111
LED:  1 count:       112
LED:  1 count:       113
LED:  1 count:       114
LED:  1 count:       115
LED:  1 count:       116
LED:  1 count:       117
LED:  1 count:       118
LED:  1 count:       119
LED:  1 count:       120
LED:  1 count:       121
LED:  1 count:       122
LED:  1 count:       123
LED:  1 count:       124
LED:  1 count:       125
LED:  1 count:       126
LED:  1 count:       127
LED:  1 count:       128
LED:  1 count:       129
LED:  1 count:       130
LED:  1 count:       131
LED:  1 count:       132
LED:  1 count:       133
LED:  1 count:       134
LED:  1 count:       135
LED:  1 count:       136
LED:  1 count:       137
LED:  1 count:       138
LED:  1 count:       139
LED:  1 count:       140
LED:  1 count:       141
LED:  1 count:       142
LED:  1 count:       143
LED:  1 count:       144
LED:  1 count:       145
LED:  1 count:       146
LED:  1 count:       147
LED:  1 count:       148
LED:  1 count:       149
LED:  1 count:       150
LED:  1 count:       151
LED:  1 count:       152
LED:  1 count:       153
LED:  1 count:       154
LED:  1 count:       155
LED:  1 count:       156
LED:  1 count:       157
LED:  1 count:       158
LED:  1 count:       159
LED:  1 count:       160
LED:  1 count:       161
LED:  1 count:       162
LED:  1 count:       163
LED:  1 count:       164
LED:  1 count:       165
LED:  1 count:       166
LED:  1 count:       167
LED:  1 count:       168
LED:  1 count:       169
LED:  1 count:       170
LED:  1 count:       171
LED:  1 count:       172
LED:  1 count:       173
LED:  1 count:       174
LED:  1 count:       175
LED:  1 count:       176
LED:  1 count:       177
LED:  1 count:       178
LED:  1 count:       179
LED:  1 count:       180
LED:  1 count:       181
LED:  1 count:       182
LED:  1 count:       183
LED:  1 count:       184
LED:  1 count:       185
LED:  1 count:       186
LED:  1 count:       187
LED:  1 count:       188
LED:  1 count:       189
LED:  1 count:       190
LED:  1 count:       191
LED:  1 count:       192
LED:  1 count:       193
LED:  1 count:       194
LED:  1 count:       195
LED:  1 count:       196
LED:  1 count:       197
LED:  1 count:       198
LED:  1 count:       199
LED:  1 count:       200
LED:  1 count:       201
LED:  1 count:       202
LED:  1 count:       203
LED:  1 count:       204
LED:  1 count:       205
LED:  1 count:       206
LED:  1 count:       207
LED:  1 count:       208
LED:  1 count:       209
LED:  1 count:       210
LED:  1 count:       211
LED:  1 count:       212
LED:  1 count:       213
LED:  1 count:       214
LED:  1 count:       215
LED:  1 count:       216
LED:  1 count:       217
LED:  1 count:       218
LED:  1 count:       219
LED:  1 count:       220
LED:  1 count:       221
LED:  1 count:       222
LED:  1 count:       223
LED:  1 count:       224
LED:  1 count:       225
LED:  1 count:       226
LED:  1 count:       227
LED:  1 count:       228
LED:  1 count:       229
LED:  1 count:       230
LED:  1 count:       231
LED:  1 count:       232
LED:  1 count:       233
LED:  1 count:       234
LED:  1 count:       235
LED:  1 count:       236
LED:  1 count:       237
LED:  1 count:       238
LED:  1 count:       239
LED:  1 count:       240
LED:  1 count:       241
LED:  1 count:       242
LED:  1 count:       243
LED:  1 count:       244
LED:  1 count:       245
LED:  1 count:       246
LED:  1 count:       247
LED:  1 count:       248
LED:  1 count:       249
LED:  1 count:       250
LED:  1 count:       251
LED:  1 count:       252
LED:  1 count:       253
LED:  1 count:       254
LED:  1 count:       255
LED:  1 count:       256
LED:  1 count:       257
LED:  1 count:       258
LED:  1 count:       259
LED:  1 count:       260
LED:  1 count:       261
LED:  1 count:       262
LED:  1 count:       263
LED:  1 count:       264
LED:  1 count:       265
LED:  1 count:       266
LED:  1 count:       267
LED:  1 count:       268
LED:  1 count:       269
LED:  1 count:       270
LED:  1 count:       271
LED:  1 count:       272
LED:  1 count:       273
LED:  1 count:       274
LED:  1 count:       275
LED:  1 count:       276
LED:  1 count:       277
LED:  1 count:       278
LED:  1 count:       279
LED:  1 count:       280
LED:  1 count:       281
LED:  1 count:       282
LED:  1 count:       283
LED:  1 count:       284
LED:  1 count:       285
LED:  1 count:       286
LED:  1 count:       287
LED:  1 count:       288
LED:  1 count:       289
LED:  1 count:       290
LED:  1 count:       291
LED:  1 count:       292
LED:  1 count:       293
LED:  1 count:       294
LED:  1 count:       295
LED:  1 count:       296
LED:  1 count:       297
LED:  1 count:       298
LED:  1 count:       299
LED:  1 count:       300
LED:  1 count:       301
LED:  1 count:       302
LED:  1 count:       303
LED:  1 count:       304
LED:  1 count:       305
LED:  1 count:       306
LED:  1 count:       307
LED:  1 count:       308
LED:  1 count:       309
LED:  1 count:       310
LED:  1 count:       311
LED:  1 count:       312
LED:  1 count:       313
LED:  1 count:       314
LED:  1 count:       315
LED:  1 count:       316
LED:  1 count:       317
LED:  1 count:       318
LED:  1 count:       319
LED:  1 count:       320
LED:  1 count:       321
LED:  1 count:       322
LED:  1 count:       323
LED:  1 count:       324
LED:  1 count:       325
LED:  1 count:       326
LED:  1 count:       327
LED:  1 count:       328
LED:  1 count:       329
LED:  1 count:       330
LED:  1 count:       331
LED:  1 count:       332
LED:  1 count:       333
LED:  1 count:       334
LED:  1 count:       335
LED:  1 count:       336
LED:  1 count:       337
LED:  1 count:       338
LED:  1 count:       339
LED:  1 count:       340
LED:  1 count:       341
LED:  1 count:       342
LED:  1 count:       343
LED:  1 count:       344
LED:  1 count:       345
LED:  1 count:       346
LED:  1 count:       347
LED:  1 count:       348
LED:  1 count:       349
LED:  1 count:       350
LED:  1 count:       351
LED:  1 count:       352
LED:  1 count:       353
LED:  1 count:       354
LED:  1 count:       355
LED:  1 count:       356
LED:  1 count:       357
LED:  1 count:       358
LED:  1 count:       359
LED:  1 count:       360
LED:  1 count:       361
LED:  1 count:       362
LED:  1 count:       363
LED:  1 count:       364
LED:  1 count:       365
LED:  1 count:       366
LED:  1 count:       367
LED:  1 count:       368
LED:  1 count:       369
LED:  1 count:       370
LED:  1 count:       371
LED:  1 count:       372
LED:  1 count:       373
LED:  1 count:       374
LED:  1 count:       375
LED:  1 count:       376
LED:  1 count:       377
LED:  1 count:       378
LED:  1 count:       379
LED:  1 count:       380
LED:  1 count:       381
LED:  1 count:       382
LED:  1 count:       383
LED:  1 count:       384
LED:  1 count:       385
LED:  1 count:       386
LED:  1 count:       387
LED:  1 count:       388
LED:  1 count:       389
LED:  1 count:       390
LED:  1 count:       391
LED:  1 count:       392
LED:  1 count:       393
LED:  1 count:       394
LED:  1 count:       395
LED:  1 count:       396
LED:  1 count:       397
LED:  1 count:       398
LED:  1 count:       399
LED:  1 count:       400
LED:  1 count:       401
LED:  1 count:       402
LED:  1 count:       403
LED:  1 count:       404
LED:  1 count:       405
LED:  1 count:       406
LED:  1 count:       407
LED:  1 count:       408
LED:  1 count:       409
LED:  1 count:       410
LED:  1 count:       411
LED:  1 count:       412
LED:  1 count:       413
LED:  1 count:       414
LED:  1 count:       415
LED:  1 count:       416
LED:  1 count:       417
LED:  1 count:       418
LED:  1 count:       419
LED:  1 count:       420
LED:  1 count:       421
LED:  1 count:       422
LED:  1 count:       423
LED:  1 count:       424
LED:  1 count:       425
LED:  1 count:       426
LED:  1 count:       427
LED:  1 count:       428
LED:  1 count:       429
LED:  1 count:       430
LED:  1 count:       431
LED:  1 count:       432
LED:  1 count:       433
LED:  1 count:       434
LED:  1 count:       435
LED:  1 count:       436
LED:  1 count:       437
LED:  1 count:       438
LED:  1 count:       439
LED:  1 count:       440
LED:  1 count:       441
LED:  1 count:       442
LED:  1 count:       443
LED:  1 count:       444
LED:  1 count:       445
LED:  1 count:       446
LED:  1 count:       447
LED:  1 count:       448
LED:  1 count:       449
LED:  1 count:       450
LED:  1 count:       451
LED:  1 count:       452
LED:  1 count:       453
LED:  1 count:       454
LED:  1 count:       455
LED:  1 count:       456
LED:  1 count:       457
LED:  1 count:       458
LED:  1 count:       459
LED:  1 count:       460
LED:  1 count:       461
LED:  1 count:       462
LED:  1 count:       463
LED:  1 count:       464
LED:  1 count:       465
LED:  1 count:       466
LED:  1 count:       467
LED:  1 count:       468
LED:  1 count:       469
LED:  1 count:       470
LED:  1 count:       471
LED:  1 count:       472
LED:  1 count:       473
LED:  1 count:       474
LED:  1 count:       475
LED:  1 count:       476
LED:  1 count:       477
LED:  1 count:       478
LED:  1 count:       479
LED:  1 count:       480
LED:  1 count:       481
LED:  1 count:       482
LED:  1 count:       483
LED:  1 count:       484
LED:  1 count:       485
LED:  1 count:       486
LED:  1 count:       487
LED:  1 count:       488
LED:  1 count:       489
LED:  1 count:       490
LED:  1 count:       491
LED:  1 count:       492
LED:  1 count:       493
LED:  1 count:       494
LED:  1 count:       495
LED:  1 count:       496
LED:  1 count:       497
LED:  1 count:       498
LED:  1 count:       499
LED:  1 count:       500
LED:  1 count:       501
LED:  1 count:       502
LED:  1 count:       503
LED:  1 count:       504
LED:  1 count:       505
LED:  1 count:       506
LED:  1 count:       507
LED:  1 count:       508
LED:  1 count:       509
LED:  1 count:       510
LED:  1 count:       511
LED:  1 count:       512
LED:  1 count:       513
LED:  1 count:       514
LED:  1 count:       515
LED:  1 count:       516
LED:  1 count:       517
LED:  1 count:       518
LED:  1 count:       519
LED:  1 count:       520
LED:  1 count:       521
LED:  1 count:       522
LED:  1 count:       523
LED:  1 count:       524
LED:  1 count:       525
LED:  1 count:       526
LED:  1 count:       527
LED:  1 count:       528
LED:  1 count:       529
LED:  1 count:       530
LED:  1 count:       531
LED:  1 count:       532
LED:  1 count:       533
LED:  1 count:       534
LED:  1 count:       535
LED:  1 count:       536
LED:  1 count:       537
LED:  1 count:       538
LED:  1 count:       539
LED:  1 count:       540
LED:  1 count:       541
LED:  1 count:       542
LED:  1 count:       543
LED:  1 count:       544
LED:  1 count:       545
LED:  1 count:       546
LED:  1 count:       547
LED:  1 count:       548
LED:  1 count:       549
LED:  1 count:       550
LED:  1 count:       551
LED:  1 count:       552
LED:  1 count:       553
LED:  1 count:       554
LED:  1 count:       555
LED:  1 count:       556
LED:  1 count:       557
LED:  1 count:       558
LED:  1 count:       559
LED:  1 count:       560
LED:  1 count:       561
LED:  1 count:       562
LED:  1 count:       563
LED:  1 count:       564
LED:  1 count:       565
LED:  1 count:       566
LED:  1 count:       567
LED:  1 count:       568
LED:  1 count:       569
LED:  1 count:       570
LED:  1 count:       571
LED:  1 count:       572
LED:  1 count:       573
LED:  1 count:       574
LED:  1 count:       575
LED:  1 count:       576
LED:  1 count:       577
LED:  1 count:       578
LED:  1 count:       579
LED:  1 count:       580
LED:  1 count:       581
LED:  1 count:       582
LED:  1 count:       583
LED:  1 count:       584
LED:  1 count:       585
LED:  1 count:       586
LED:  1 count:       587
LED:  1 count:       588
LED:  1 count:       589
LED:  1 count:       590
LED:  1 count:       591
LED:  1 count:       592
LED:  1 count:       593
LED:  1 count:       594
LED:  1 count:       595
LED:  1 count:       596
LED:  1 count:       597
LED:  1 count:       598
LED:  1 count:       599
LED:  1 count:       600
LED:  1 count:       601
LED:  1 count:       602
LED:  1 count:       603
LED:  1 count:       604
LED:  1 count:       605
LED:  1 count:       606
LED:  1 count:       607
LED:  1 count:       608
LED:  1 count:       609
LED:  1 count:       610
LED:  1 count:       611
LED:  1 count:       612
LED:  1 count:       613
LED:  1 count:       614
LED:  1 count:       615
LED:  1 count:       616
LED:  1 count:       617
LED:  1 count:       618
LED:  1 count:       619
LED:  1 count:       620
LED:  1 count:       621
LED:  1 count:       622
LED:  1 count:       623
LED:  1 count:       624
LED:  1 count:       625
LED:  1 count:       626
LED:  1 count:       627
LED:  1 count:       628
LED:  1 count:       629
LED:  1 count:       630
LED:  1 count:       631
LED:  1 count:       632
LED:  1 count:       633
LED:  1 count:       634
LED:  1 count:       635
LED:  1 count:       636
LED:  1 count:       637
LED:  1 count:       638
LED:  1 count:       639
LED:  1 count:       640
LED:  1 count:       641
LED:  1 count:       642
LED:  1 count:       643
LED:  1 count:       644
LED:  1 count:       645
LED:  1 count:       646
LED:  1 count:       647
LED:  1 count:       648
LED:  1 count:       649
LED:  1 count:       650
LED:  1 count:       651
LED:  1 count:       652
LED:  1 count:       653
LED:  1 count:       654
LED:  1 count:       655
LED:  1 count:       656
LED:  1 count:       657
LED:  1 count:       658
LED:  1 count:       659
LED:  1 count:       660
LED:  1 count:       661
LED:  1 count:       662
LED:  1 count:       663
LED:  1 count:       664
LED:  1 count:       665
LED:  1 count:       666
LED:  1 count:       667
LED:  1 count:       668
LED:  1 count:       669
LED:  1 count:       670
LED:  1 count:       671
LED:  1 count:       672
LED:  1 count:       673
LED:  1 count:       674
LED:  1 count:       675
LED:  1 count:       676
LED:  1 count:       677
LED:  1 count:       678
LED:  1 count:       679
LED:  1 count:       680
LED:  1 count:       681
LED:  1 count:       682
LED:  1 count:       683
LED:  1 count:       684
LED:  1 count:       685
LED:  1 count:       686
LED:  1 count:       687
LED:  1 count:       688
LED:  1 count:       689
LED:  1 count:       690
LED:  1 count:       691
LED:  1 count:       692
LED:  1 count:       693
LED:  1 count:       694
LED:  1 count:       695
LED:  1 count:       696
LED:  1 count:       697
LED:  1 count:       698
LED:  1 count:       699
LED:  1 count:       700
LED:  1 count:       701
LED:  1 count:       702
LED:  1 count:       703
LED:  1 count:       704
LED:  1 count:       705
LED:  1 count:       706
LED:  1 count:       707
LED:  1 count:       708
LED:  1 count:       709
LED:  1 count:       710
LED:  1 count:       711
LED:  1 count:       712
LED:  1 count:       713
LED:  1 count:       714
LED:  1 count:       715
LED:  1 count:       716
LED:  1 count:       717
LED:  1 count:       718
LED:  1 count:       719
LED:  1 count:       720
LED:  1 count:       721
LED:  1 count:       722
LED:  1 count:       723
LED:  1 count:       724
LED:  1 count:       725
LED:  1 count:       726
LED:  1 count:       727
LED:  1 count:       728
LED:  1 count:       729
LED:  1 count:       730
LED:  1 count:       731
LED:  1 count:       732
LED:  1 count:       733
LED:  1 count:       734
LED:  1 count:       735
LED:  1 count:       736
LED:  1 count:       737
LED:  1 count:       738
LED:  1 count:       739
LED:  1 count:       740
LED:  1 count:       741
LED:  1 count:       742
LED:  1 count:       743
LED:  1 count:       744
LED:  1 count:       745
LED:  1 count:       746
LED:  1 count:       747
LED:  1 count:       748
LED:  1 count:       749
LED:  1 count:       750
LED:  1 count:       751
LED:  1 count:       752
LED:  1 count:       753
LED:  1 count:       754
LED:  1 count:       755
LED:  1 count:       756
LED:  1 count:       757
LED:  1 count:       758
LED:  1 count:       759
LED:  1 count:       760
LED:  1 count:       761
LED:  1 count:       762
LED:  1 count:       763
LED:  1 count:       764
LED:  1 count:       765
LED:  1 count:       766
LED:  1 count:       767
LED:  1 count:       768
LED:  1 count:       769
LED:  1 count:       770
LED:  1 count:       771
LED:  1 count:       772
LED:  1 count:       773
LED:  1 count:       774
LED:  1 count:       775
LED:  1 count:       776
LED:  1 count:       777
LED:  1 count:       778
LED:  1 count:       779
LED:  1 count:       780
LED:  1 count:       781
LED:  1 count:       782
LED:  1 count:       783
LED:  1 count:       784
LED:  1 count:       785
LED:  1 count:       786
LED:  1 count:       787
LED:  1 count:       788
LED:  1 count:       789
LED:  1 count:       790
LED:  1 count:       791
LED:  1 count:       792
LED:  1 count:       793
LED:  1 count:       794
LED:  1 count:       795
LED:  1 count:       796
LED:  1 count:       797
LED:  1 count:       798
LED:  1 count:       799
LED:  1 count:       800
LED:  1 count:       801
LED:  1 count:       802
LED:  1 count:       803
LED:  1 count:       804
LED:  1 count:       805
LED:  1 count:       806
LED:  1 count:       807
LED:  1 count:       808
LED:  1 count:       809
LED:  1 count:       810
LED:  1 count:       811
LED:  1 count:       812
LED:  1 count:       813
LED:  1 count:       814
LED:  1 count:       815
LED:  1 count:       816
LED:  1 count:       817
LED:  1 count:       818
LED:  1 count:       819
LED:  1 count:       820
LED:  1 count:       821
LED:  1 count:       822
LED:  1 count:       823
LED:  1 count:       824
LED:  1 count:       825
LED:  1 count:       826
LED:  1 count:       827
LED:  1 count:       828
LED:  1 count:       829
LED:  1 count:       830
LED:  1 count:       831
LED:  1 count:       832
LED:  1 count:       833
LED:  1 count:       834
LED:  1 count:       835
LED:  1 count:       836
LED:  1 count:       837
LED:  1 count:       838
LED:  1 count:       839
LED:  1 count:       840
LED:  1 count:       841
LED:  1 count:       842
LED:  1 count:       843
LED:  1 count:       844
LED:  1 count:       845
LED:  1 count:       846
LED:  1 count:       847
LED:  1 count:       848
LED:  1 count:       849
LED:  1 count:       850
LED:  1 count:       851
LED:  1 count:       852
LED:  1 count:       853
LED:  1 count:       854
LED:  1 count:       855
LED:  1 count:       856
LED:  1 count:       857
LED:  1 count:       858
LED:  1 count:       859
LED:  1 count:       860
LED:  1 count:       861
LED:  1 count:       862
LED:  1 count:       863
LED:  1 count:       864
LED:  1 count:       865
LED:  1 count:       866
LED:  1 count:       867
LED:  1 count:       868
LED:  1 count:       869
LED:  1 count:       870
LED:  1 count:       871
LED:  1 count:       872
LED:  1 count:       873
LED:  1 count:       874
LED:  1 count:       875
LED:  1 count:       876
LED:  1 count:       877
LED:  1 count:       878
LED:  1 count:       879
LED:  1 count:       880
LED:  1 count:       881
LED:  1 count:       882
LED:  1 count:       883
LED:  1 count:       884
LED:  1 count:       885
LED:  1 count:       886
LED:  1 count:       887
LED:  1 count:       888
LED:  1 count:       889
LED:  1 count:       890
LED:  1 count:       891
LED:  1 count:       892
LED:  1 count:       893
LED:  1 count:       894
LED:  1 count:       895
LED:  1 count:       896
LED:  1 count:       897
LED:  1 count:       898
LED:  1 count:       899
LED:  1 count:       900
LED:  1 count:       901
LED:  1 count:       902
LED:  1 count:       903
LED:  1 count:       904
LED:  1 count:       905
LED:  1 count:       906
LED:  1 count:       907
LED:  1 count:       908
LED:  1 count:       909
LED:  1 count:       910
LED:  1 count:       911
LED:  1 count:       912
LED:  1 count:       913
LED:  1 count:       914
LED:  1 count:       915
LED:  1 count:       916
LED:  1 count:       917
LED:  1 count:       918
LED:  1 count:       919
LED:  1 count:       920
LED:  1 count:       921
LED:  1 count:       922
LED:  1 count:       923
LED:  1 count:       924
LED:  1 count:       925
LED:  1 count:       926
LED:  1 count:       927
LED:  1 count:       928
LED:  1 count:       929
LED:  1 count:       930
LED:  1 count:       931
LED:  1 count:       932
LED:  1 count:       933
LED:  1 count:       934
LED:  1 count:       935
LED:  1 count:       936
LED:  1 count:       937
LED:  1 count:       938
LED:  1 count:       939
LED:  1 count:       940
LED:  1 count:       941
LED:  1 count:       942
LED:  1 count:       943
LED:  1 count:       944
LED:  1 count:       945
LED:  1 count:       946
LED:  1 count:       947
LED:  1 count:       948
LED:  1 count:       949
LED:  1 count:       950
LED:  1 count:       951
LED:  1 count:       952
LED:  1 count:       953
LED:  1 count:       954
LED:  1 count:       955
"""


def test():
    veriloggen.reset()
    test_module = simulation_simulator_verilator.mkTest()
    sim = simulation.Simulator(test_module, sim='verilator')
    rslt = sim.run(outputfile='verilator.out', sim_time=1000 * 20)
    assert(expected_rslt == rslt)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    filename = os.path.dirname(os.path.abspath(__file__)) + '/led.v'
    modules = from_verilog.read_verilog_module(filename)
    m = modules['blinkled']
    
    # change the module name
    m.name = 'modified_led'
    
    # add new statements
    enable = m.Input('enable')
    busy = m.Output('busy')

    # get current ports
    ports = m.get_ports()
    led = ports['LED']

    # modify the data width
    led.width = 32
    
    old_statement = m.always[0].statement[0].false_statement
    m.always[0].statement[0].false_statement = If(enable)(*old_statement)
    m.Assign( busy(m.variable['count'] < 1023) )
    
    return m

def mkTop():
    m = Module('top')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', width)

    led = mkLed()
    params = m.copy_params(led)
    ports = m.copy_ports(led)
    
    m.Instance(led, 'inst_blinkled', params, ports)

    return m

if __name__ == '__main__':
    top = mkTop()
    verilog = top.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function


class Allocator(object):

    def __init__(self, **custom_methods):
        self.custom_methods = custom_methods

    def allocate(self, m, seq, nodes):
        for node in sorted(nodes, key=lambda x: (-1, x.object_id) if x.start_stage is None else
                           (x.start_stage, x.object_id)):
            self.implement(m, seq, node)

    def implement(self, m, seq, node):
        if node.__class__.__name__ in self.custom_methods:
            self.custom_methods[node.__class__.__name__](node)
            return

        if hasattr(node, '_implement'):
            node._implement(m, seq)
            return

        raise TypeError(
            "Not found implement() method for Type '%s'" % str(type(node)))



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_saturate_add

expected_verilog = """
module test;

  reg uut_CLK;
  reg uut_RST;
  wire [8-1:0] uut_LED;

  blinkled
  uut
  (
    .CLK(uut_CLK),
    .RST(uut_RST),
    .LED(uut_LED)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, uut_CLK, uut_RST, uut_LED);
  end


  initial begin
    uut_CLK = 0;
    forever begin
      #5 uut_CLK = !uut_CLK;
    end
  end


  initial begin
    uut_RST = 0;
    #100;
    uut_RST = 1;
    #100;
    uut_RST = 0;
    #100000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg [10-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
      LED <= 0;
    end else begin
      count <= (count + 1 > 1023)? 1023 : count + 1;
      if(count == 1023) begin
        LED <= (LED + 'sd1 > 255)? 255 : LED + 'sd1;
      end 
      $display("LED:%d count:%d", LED, count);
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = types_saturate_add.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)
